import os
from importlib import import_module

from dependency_injector import containers, providers

from isar.apis.api import API
from isar.apis.robot_control.robot_controller import RobotController
from isar.apis.schedule.scheduling_controller import SchedulingController
from isar.apis.security.authentication import Authenticator
from isar.config.keyvault.keyvault_service import Keyvault
from isar.config.settings import settings
from isar.mission_planner.local_planner import LocalPlanner
from isar.mission_planner.sequential_task_selector import SequentialTaskSelector
from isar.mission_planner.task_selector_interface import TaskSelectorInterface
from isar.models.events import Events, SharedState
from isar.robot.robot import Robot
from isar.services.utilities.robot_utilities import RobotUtilities
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.state_machine import StateMachine
from isar.storage.blob_storage import BlobStorage
from isar.storage.local_storage import LocalStorage
from isar.storage.uploader import Uploader
from robot_interface.telemetry.mqtt_client import MqttPublisher


class ApplicationContainer(containers.DeclarativeContainer):
    config = providers.Configuration(pydantic_settings=[settings])

    # Core services
    keyvault = providers.Singleton(
        Keyvault,
        keyvault_name=settings.KEYVAULT_NAME,
        client_id=settings.AZURE_CLIENT_ID,
        client_secret=os.environ.get("AZURE_CLIENT_SECRET"),
        tenant_id=settings.AZURE_TENANT_ID,
    )

    # Events and shared state
    events = providers.Singleton(Events)
    shared_state = providers.Singleton(SharedState)

    # Robot-related services
    robot_interface = providers.Singleton(
        lambda: import_module(f"{settings.ROBOT_PACKAGE}.robotinterface").Robot()
    )
    robot_utilities = providers.Singleton(RobotUtilities, robot=robot_interface)

    # Mqtt client
    mqtt_client = (
        providers.Singleton(
            MqttPublisher,
            mqtt_queue=providers.Callable(events.provided.mqtt_queue),
        )
        if settings.MQTT_ENABLED
        else None
    )

    # API and controllers
    authenticator = providers.Singleton(Authenticator)
    scheduling_utilities = providers.Singleton(
        SchedulingUtilities,
        events=events,
        shared_state=shared_state,
        mission_planner=providers.Singleton(LocalPlanner),
    )
    scheduling_controller = providers.Singleton(
        SchedulingController, scheduling_utilities=scheduling_utilities
    )
    robot_controller = providers.Singleton(
        RobotController, robot_utilities=robot_utilities
    )
    api = providers.Singleton(
        API,
        authenticator=authenticator,
        scheduling_controller=scheduling_controller,
        robot_controller=robot_controller,
        keyvault=keyvault,
        mqtt_publisher=mqtt_client,
    )

    # Storage
    storage_handlers_temp = []
    if settings.STORAGE_LOCAL_ENABLED:
        local_storage = providers.Singleton(LocalStorage)
        storage_handlers_temp.append(local_storage)
    if settings.STORAGE_BLOB_ENABLED:
        blob_storage = providers.Singleton(BlobStorage, keyvault=keyvault)
        storage_handlers_temp.append(blob_storage)
    storage_handlers = providers.List(*storage_handlers_temp)

    # State machine
    task_selector = providers.Singleton(
        SequentialTaskSelector
        if settings.TASK_SELECTOR == "sequential"
        else TaskSelectorInterface
    )

    state_machine = providers.Singleton(
        StateMachine,
        events=events,
        shared_state=shared_state,
        robot=robot_interface,
        mqtt_publisher=mqtt_client,
        task_selector=task_selector,
    )

    # Robot
    robot = providers.Singleton(
        Robot, events=events, robot=robot_interface, shared_state=shared_state
    )

    # Uploader
    uploader = providers.Singleton(
        Uploader,
        events=events,
        storage_handlers=storage_handlers,
        mqtt_publisher=mqtt_client,
    )


def get_injector() -> ApplicationContainer:
    container = ApplicationContainer()
    container.init_resources()
    container.wire(modules=[__name__])
    container.config.from_dict(
        {
            "KEYVAULT_NAME": settings.KEYVAULT_NAME,
            "MQTT_ENABLED": settings.MQTT_ENABLED,
            "TASK_SELECTOR": settings.TASK_SELECTOR,
        }
    )

    print("Loaded the following module configurations:")
    for provider_name, provider in container.providers.items():
        provider_repr = repr(provider)
        simplified_provider = provider_repr.split(".")[-1].split(">")[0]
        print(f"    {provider_name:<20}: {simplified_provider}")

    return container



import json
import time
from copy import deepcopy
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import List

import numpy as np
from alitra import Frame, Position
from fastapi import FastAPI
from starlette.testclient import TestClient

from isar.apis.api import API
from isar.state_machine.states_enum import States
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.task import ReturnToHome


def test_successful_mission(
    injector_turtlebot, state_machine_thread, uploader_thread, access_token
) -> None:
    integration_test_timeout: timedelta = timedelta(minutes=5)
    app: FastAPI = injector_turtlebot.get(API).get_app()
    client: TestClient = TestClient(app=app)

    _id: int = 2

    _ = client.post(
        f"schedule/start-mission?ID={_id}",
        headers={"Authorization": "Bearer {}".format(access_token)},
    )
    time.sleep(5)

    mission_id: str = state_machine_thread.state_machine.current_mission.id
    mission: Mission = deepcopy(state_machine_thread.state_machine.current_mission)

    start_time: datetime = datetime.now(timezone.utc)
    while state_machine_thread.state_machine.current_state != States.RobotStandingStill:
        if (datetime.now(timezone.utc) - start_time) > integration_test_timeout:
            raise TimeoutError
        time.sleep(5)

    mission_result_folder: Path = Path(f"tests/results/{mission_id}")

    drive_to_tasks: List[ReturnToHome] = [
        task for task in mission.tasks if isinstance(task, ReturnToHome)
    ]

    expected_positions: list = [task.pose for task in drive_to_tasks]

    paths = mission_result_folder.rglob("*.json")
    for path in paths:
        with open(path) as json_file:
            metadata = json.load(json_file)
        files_metadata: dict = metadata["data"][0]["files"][0]
        filename: str = files_metadata["file_name"]
        inspection_file: Path = mission_result_folder.joinpath(filename)

        actual_position: Position = Position(
            x=files_metadata["x"],
            y=files_metadata["y"],
            z=files_metadata["z"],
            frame=Frame("asset"),
        )

        close_to_expected_positions: List[bool] = []
        for expected_position in expected_positions:
            close_to_expected_positions.append(
                np.allclose(
                    actual_position.to_list(), expected_position.to_list(), atol=0.2
                )
            )

        assert any(close_to_expected_positions)
        assert inspection_file.exists()
    assert (
        len(list(mission_result_folder.glob("*"))) == 6
    )  # Expected two images, one thermal + three metadata files



import os
from pathlib import Path
from typing import List

from injector import Injector
from transitions import State
from transitions.extensions import GraphMachine

from isar.modules import get_injector
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.transitions.mission import get_mission_transitions
from isar.state_machine.transitions.return_home import get_return_home_transitions
from isar.state_machine.transitions.robot_status import get_robot_status_transitions


def extract_function_name_from_callable(func):
    try:
        closure = getattr(func, "__closure__", None)
        if closure:
            for cell in closure:
                obj = cell.cell_contents
                if callable(obj):
                    return getattr(obj, "__name__", str(obj))

        return getattr(func, "__name__", str(func))
    except Exception:
        return "unknown"


def embed_conditions_in_trigger(transitions: List[dict]) -> List[dict]:
    updated = []
    for t in transitions:
        trigger = t.get("trigger", "")
        conditions = t.get("conditions", [])

        if not isinstance(conditions, list):
            conditions = [conditions]

        condition_names = []

        for cond in conditions:
            name = extract_function_name_from_callable(cond)
            condition_names.append(name)

        befores = t.get("before", [])
        if not isinstance(befores, list):
            befores = [befores]
        before_names = [extract_function_name_from_callable(b) for b in befores]

        label_parts = [trigger]
        if condition_names:
            label_parts.append(f"[{'; '.join(condition_names)}]")
        if before_names:
            label_parts.append(f"/ {'; '.join(before_names)}")

        new_trigger = " ".join(label_parts)

        t_copy = t.copy()
        t_copy["trigger"] = new_trigger
        updated.append(t_copy)

    return updated


def draw_diagram(states: List[State], transitions: List[dict], name: str):
    transitions_with_conditions = embed_conditions_in_trigger(transitions)
    machine = GraphMachine(states=states, initial="unknown_status", queued=True)
    machine.add_transitions(transitions_with_conditions)
    gp = machine.get_combined_graph()

    state_machine_diagram_file = (
        Path(__file__).parent.resolve().joinpath(Path(f"{name}.png"))
    )

    if os.path.isfile(state_machine_diagram_file):
        os.remove(state_machine_diagram_file)

    gp.draw(str(state_machine_diagram_file), prog="dot", format="png")


if __name__ == "__main__":
    injector: Injector = get_injector()
    state_machine: StateMachine = injector.state_machine()

    mission_extended_transitions: List[dict] = []
    for transition in get_mission_transitions(state_machine):
        mission_extended_transitions.append(transition)

    for transition in get_return_home_transitions(state_machine):
        mission_extended_transitions.append(transition)

    draw_diagram(
        states=state_machine.states,
        transitions=mission_extended_transitions,
        name="mission_state_machine_diagram",
    )

    draw_diagram(
        states=state_machine.states,
        transitions=get_robot_status_transitions(state_machine),
        name="robot_status_state_machine_diagram",
    )

    draw_diagram(
        states=state_machine.states,
        transitions=state_machine.transitions,
        name="full_state_machine_diagram",
    )



import logging
import logging.config
from importlib.resources import as_file, files

import yaml
from uvicorn.logging import ColourizedFormatter

from isar.config.settings import settings


def setup_loggers() -> None:
    log_levels: dict = settings.LOG_LEVELS
    log_config = load_log_config()

    logging.config.dictConfig(log_config)

    handlers = []
    if settings.LOG_HANDLER_LOCAL_ENABLED:
        handlers.append(configure_console_handler(log_config=log_config))

    for log_handler in handlers:
        for loggers in log_config["loggers"].keys():
            logging.getLogger(loggers).addHandler(log_handler)
            logging.getLogger(loggers).setLevel(log_levels[loggers])
        logging.getLogger().addHandler(log_handler)


def load_log_config():
    source = files("isar").joinpath("config").joinpath("logging.conf")
    with as_file(source) as f:
        log_config = yaml.safe_load(open(f))
    return log_config


def configure_console_handler(log_config: dict) -> logging.Handler:
    handler = logging.StreamHandler()
    handler.setLevel(log_config["root"]["level"])
    handler.setFormatter(
        ColourizedFormatter(
            log_config["formatters"]["colourized"]["format"],
            style="{",
            use_colors=True,
        )
    )
    return handler



from typing import TYPE_CHECKING, Callable

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


def def_transition(
    state_machine: "StateMachine", transition_function: Callable[["StateMachine"], bool]
) -> Callable[[], bool]:
    return lambda: transition_function(state_machine)



from collections import deque
from queue import Empty, Queue
from typing import Generic, Optional, TypeVar

from transitions import State

from isar.apis.models.models import ControlMissionResponse, MissionStartResponse
from isar.config.settings import settings
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import RobotStatus, TaskStatus
from robot_interface.models.mission.task import TASKS

T = TypeVar("T")
T1 = TypeVar("T1")
T2 = TypeVar("T2")


class Event(Queue[T]):
    def __init__(self) -> None:
        super().__init__(maxsize=1)

    def trigger_event(self, data: T, timeout: int = None) -> None:
        try:
            self.put(data, timeout=timeout)
        except Exception:
            if timeout is not None:
                raise EventTimeoutError
            return None

    def consume_event(self, timeout: int = None) -> Optional[T]:
        try:
            return self.get(block=timeout is not None, timeout=timeout)
        except Empty:
            if timeout is not None:
                raise EventTimeoutError
            return None

    def clear_event(self) -> None:
        while True:
            try:
                self.get(block=False)
            except Empty:
                break

    def has_event(self) -> bool:
        return (
            self.qsize() != 0
        )  # Queue size is not reliable, but should be sufficient for this case

    def check(self) -> Optional[T]:
        if not self._qsize():
            return None
        with self.mutex:
            queueList = list(self.queue)
            return queueList.pop()

    def update(self, item: T):
        with self.mutex:
            self.queue: deque[T] = deque()
            self.queue.append(item)


class Events:
    def __init__(self) -> None:
        self.api_requests: APIRequests = APIRequests()
        self.state_machine_events: StateMachineEvents = StateMachineEvents()
        self.robot_service_events: RobotServiceEvents = RobotServiceEvents()

        self.upload_queue: Queue = Queue(maxsize=10)

        if settings.MQTT_ENABLED:
            self.mqtt_queue: Queue = Queue()


class APIEvent(Generic[T1, T2]):
    """
    Creates request and response event. The events are defined such that the request is from
    api to state machine while the response is from state machine to api.
    """

    def __init__(self):
        self.request: Event[T1] = Event()
        self.response: Event[T2] = Event()


class APIRequests:
    def __init__(self) -> None:
        self.start_mission: APIEvent[Mission, MissionStartResponse] = APIEvent()
        self.stop_mission: APIEvent[str, ControlMissionResponse] = APIEvent()
        self.pause_mission: APIEvent[bool, ControlMissionResponse] = APIEvent()
        self.resume_mission: APIEvent[bool, ControlMissionResponse] = APIEvent()
        self.return_home: APIEvent[bool, bool] = APIEvent()
        self.release_intervention_needed: APIEvent[bool, bool] = APIEvent()


class StateMachineEvents:
    def __init__(self) -> None:
        self.start_mission: Event[Mission] = Event()
        self.stop_mission: Event[bool] = Event()
        self.pause_mission: Event[bool] = Event()
        self.task_status_request: Event[str] = Event()


class RobotServiceEvents:
    def __init__(self) -> None:
        self.task_status_updated: Event[TaskStatus] = Event()
        self.task_status_failed: Event[ErrorMessage] = Event()
        self.mission_started: Event[bool] = Event()
        self.mission_failed: Event[ErrorMessage] = Event()
        self.robot_status_changed: Event[bool] = Event()
        self.mission_failed_to_stop: Event[ErrorMessage] = Event()
        self.mission_successfully_stopped: Event[bool] = Event()


class SharedState:
    def __init__(self) -> None:
        self.state: Event[State] = Event()
        self.robot_status: Event[RobotStatus] = Event()
        self.state_machine_current_task: Event[TASKS] = Event()
        self.robot_battery_level: Event[float] = Event()


class EventTimeoutError(Exception):
    pass


class EventConflictError(Exception):
    pass



from typing import TYPE_CHECKING

from robot_interface.models.mission.status import RobotStatus

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


def is_offline(state_machine: "StateMachine") -> bool:
    robot_status = state_machine.shared_state.robot_status.check()
    return robot_status == RobotStatus.Offline


def is_available(state_machine: "StateMachine") -> bool:
    robot_status = state_machine.shared_state.robot_status.check()
    return robot_status == RobotStatus.Available


def is_home(state_machine: "StateMachine") -> bool:
    robot_status = state_machine.shared_state.robot_status.check()
    return robot_status == RobotStatus.Home


def is_blocked_protective_stop(state_machine: "StateMachine") -> bool:
    robot_status = state_machine.shared_state.robot_status.check()
    return robot_status == RobotStatus.BlockedProtectiveStop



from typing import TYPE_CHECKING, List

from isar.state_machine.transitions.functions.fail_mission import (
    report_failed_mission_and_finalize,
)
from isar.state_machine.transitions.functions.finish_mission import finish_mission
from isar.state_machine.transitions.functions.pause import pause_mission
from isar.state_machine.transitions.functions.resume import resume_mission
from isar.state_machine.transitions.functions.start_mission import (
    acknowledge_mission,
    initialize_robot,
    prepare_state_machine_before_mission,
    set_mission_to_in_progress,
    trigger_start_mission_event,
)
from isar.state_machine.transitions.functions.stop import (
    stop_mission_cleanup,
    stop_mission_failed,
    stop_return_home_mission_cleanup,
    stop_return_home_mission_failed,
    trigger_stop_mission_event,
)
from isar.state_machine.transitions.functions.utils import def_transition

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


def get_mission_transitions(state_machine: "StateMachine") -> List[dict]:
    mission_transitions: List[dict] = [
        {
            "trigger": "pause",
            "source": state_machine.monitor_state,
            "dest": state_machine.paused_state,
            "conditions": def_transition(state_machine, pause_mission),
        },
        {
            "trigger": "pause",
            "source": state_machine.monitor_state,
            "dest": state_machine.monitor_state,
        },
        {
            "trigger": "resume",
            "source": state_machine.paused_state,
            "dest": state_machine.monitor_state,
            "conditions": def_transition(state_machine, resume_mission),
        },
        {
            "trigger": "resume",
            "source": state_machine.paused_state,
            "dest": state_machine.paused_state,
        },
        {
            "trigger": "stop",
            "source": [
                state_machine.await_next_mission_state,
                state_machine.robot_standing_still_state,
                state_machine.monitor_state,
                state_machine.returning_home_state,
                state_machine.paused_state,
            ],
            "dest": state_machine.stopping_state,
            "before": def_transition(state_machine, trigger_stop_mission_event),
        },
        {
            "trigger": "mission_stopped",
            "source": state_machine.stopping_state,
            "dest": state_machine.await_next_mission_state,
            "before": def_transition(state_machine, stop_mission_cleanup),
        },
        {
            "trigger": "mission_stopping_failed",
            "source": state_machine.stopping_state,
            "dest": state_machine.monitor_state,
            "before": def_transition(state_machine, stop_mission_failed),
        },
        {
            "trigger": "return_home_mission_stopping_failed",
            "source": state_machine.stopping_state,
            "dest": state_machine.returning_home_state,
            "before": def_transition(state_machine, stop_return_home_mission_failed),
        },
        {
            "trigger": "return_home_mission_stopped",
            "source": state_machine.stopping_state,
            "dest": state_machine.robot_standing_still_state,
            "before": def_transition(state_machine, stop_return_home_mission_cleanup),
        },
        {
            "trigger": "request_mission_start",
            "source": [
                state_machine.await_next_mission_state,
                state_machine.home_state,
                state_machine.robot_standing_still_state,
            ],
            "dest": state_machine.monitor_state,
            "prepare": def_transition(state_machine, acknowledge_mission),
            "conditions": [
                def_transition(state_machine, prepare_state_machine_before_mission),
                def_transition(state_machine, initialize_robot),
            ],
            "before": [
                def_transition(state_machine, set_mission_to_in_progress),
                def_transition(state_machine, trigger_start_mission_event),
            ],
        },
        {
            "trigger": "request_mission_start",
            "source": state_machine.await_next_mission_state,
            "dest": state_machine.await_next_mission_state,
            "before": def_transition(state_machine, report_failed_mission_and_finalize),
        },
        {
            "trigger": "request_mission_start",
            "source": state_machine.robot_standing_still_state,
            "dest": state_machine.robot_standing_still_state,
            "before": def_transition(state_machine, report_failed_mission_and_finalize),
        },
        {
            "trigger": "request_mission_start",
            "source": state_machine.home_state,
            "dest": state_machine.home_state,
            "before": def_transition(state_machine, report_failed_mission_and_finalize),
        },
        {
            "trigger": "mission_failed_to_start",
            "source": [state_machine.monitor_state, state_machine.returning_home_state],
            "dest": state_machine.robot_standing_still_state,
            "before": def_transition(state_machine, report_failed_mission_and_finalize),
        },
        {
            "trigger": "mission_finished",
            "source": state_machine.monitor_state,
            "dest": state_machine.await_next_mission_state,
            "before": def_transition(state_machine, finish_mission),
        },
    ]
    return mission_transitions



import shutil
from pathlib import Path

import pytest
from dependency_injector.wiring import providers
from fastapi.testclient import TestClient

from isar.apis.security.authentication import Authenticator
from isar.modules import ApplicationContainer
from isar.robot.robot import Robot
from isar.state_machine.state_machine import StateMachine
from isar.state_machine.states.monitor import Monitor
from isar.state_machine.states.robot_standing_still import RobotStandingStill
from isar.storage.uploader import Uploader
from tests.isar.state_machine.test_state_machine import (
    RobotServiceThreadMock,
    StateMachineThreadMock,
    UploaderThreadMock,
)
from tests.test_double.blob_storage import StorageFake
from tests.test_double.robot_interface import StubRobot


@pytest.fixture()
def container():
    """Fixture to provide the dependency-injector container without auth."""
    container = ApplicationContainer()
    container.storage_handlers.override(
        providers.List(providers.Singleton(StorageFake))
    )
    container.robot_interface.override(providers.Singleton(StubRobot))
    container.uploader.override(
        providers.Singleton(
            Uploader,
            container.events(),
            container.storage_handlers(),
            container.mqtt_client(),
        )
    )
    container.robot.override(
        providers.Singleton(
            Robot,
            events=container.events(),
            robot=container.robot_interface(),
            shared_state=container.shared_state(),
        )
    )
    return container


@pytest.fixture()
def app(container: ApplicationContainer):
    """Fixture to provide the FastAPI app."""
    container.authenticator.override(
        providers.Singleton(Authenticator, authentication_enabled=False)
    )

    app = container.api().get_app()
    return app


@pytest.fixture()
def client(app):
    """Fixture to provide a test client for the FastAPI app."""
    client = TestClient(app)
    return client


@pytest.fixture()
def client_auth(container: ApplicationContainer):
    """Fixture to provide a test client for the FastAPI app with auth."""
    container.authenticator.override(
        providers.Singleton(Authenticator, authentication_enabled=True)
    )

    app = container.api().get_app()
    client = TestClient(app)
    return client


@pytest.fixture()
def access_token():
    """Fixture to provide a dummy access token."""
    return "DummyToken"


@pytest.fixture()
def keyvault(container: ApplicationContainer):
    """Fixture to provide the Keyvault instance."""
    return container.keyvault()


@pytest.fixture()
def state_machine(container: ApplicationContainer, robot):
    """Fixture to provide the StateMachine instance."""
    return StateMachine(
        events=container.events(),
        shared_state=container.shared_state(),
        robot=robot,
        mqtt_publisher=container.mqtt_client(),
        task_selector=container.task_selector(),
    )


@pytest.fixture()
def robot_standing_still_state(state_machine):
    """Fixture to provide the Robot Standing Still state."""
    return RobotStandingStill(state_machine)


@pytest.fixture()
def monitor(state_machine):
    """Fixture to provide the Monitor state."""
    return Monitor(state_machine=state_machine)


@pytest.fixture()
def request_handler(container: ApplicationContainer):
    """Fixture to provide the RequestHandler instance."""
    return container.request_handler()


@pytest.fixture()
def local_planner(container: ApplicationContainer):
    """Fixture to provide the LocalPlanner instance."""
    return container.scheduling_utilities().mission_planner


@pytest.fixture()
def robot():
    """Fixture to provide a mock robot instance."""
    return StubRobot()


@pytest.fixture()
def scheduling_utilities(container: ApplicationContainer):
    """Fixture to provide the SchedulingUtilities instance."""
    return container.scheduling_utilities()


@pytest.fixture()
def mission_reader(container: ApplicationContainer):
    """Fixture to provide the LocalPlanner instance."""
    return container.scheduling_utilities().mission_planner


@pytest.fixture
def state_machine_thread(container: ApplicationContainer):
    state_machine_thread: StateMachineThreadMock = StateMachineThreadMock(
        container=container,
    )
    yield state_machine_thread
    state_machine_thread.join()


@pytest.fixture
def uploader_thread(container: ApplicationContainer):
    uploader_thread: UploaderThreadMock = UploaderThreadMock(container=container)
    yield uploader_thread
    uploader_thread.join()


@pytest.fixture
def robot_service_thread(container: ApplicationContainer):
    robot_service: Robot = Robot(
        events=container.events(),
        robot=container.robot_interface(),
        shared_state=container.shared_state(),
    )

    robot_service_thread: RobotServiceThreadMock = RobotServiceThreadMock(
        robot_service=robot_service
    )
    yield robot_service_thread
    robot_service_thread.join()


@pytest.fixture(autouse=True)
def run_before_and_after_tests() -> None:  # type: ignore
    results_folder: Path = Path("tests/results")
    yield

    print("Removing temporary results folder for testing")
    if results_folder.exists():
        shutil.rmtree(results_folder)
    print("Cleanup finished")



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine

from typing import TYPE_CHECKING

from isar.config.settings import settings
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import MissionStatus, RobotStatus, TaskStatus
from robot_interface.models.mission.task import ReturnToHome

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


def start_return_home_mission(state_machine: "StateMachine") -> bool:
    state_machine.start_mission(
        Mission(
            tasks=[ReturnToHome()],
            name="Return Home",
        )
    )
    return True


def should_retry_return_home(state_machine: "StateMachine") -> bool:
    if state_machine.shared_state.robot_status.check() != RobotStatus.Available:
        return False

    return (
        state_machine.returning_home_state.failed_return_home_attemps
        < settings.RETURN_HOME_RETRY_LIMIT
    )


def reset_return_home_failure_counter(state_machine: "StateMachine") -> bool:
    state_machine.returning_home_state.failed_return_home_attemps = 0
    return True


def set_return_home_status(state_machine: "StateMachine") -> bool:
    state_machine.log_mission_overview(mission=state_machine.current_mission)
    state_machine.current_mission.status = MissionStatus.InProgress

    state_machine.current_task = state_machine.task_selector.next_task()
    state_machine.current_task.status = TaskStatus.InProgress

    return True


def return_home_finished(state_machine: "StateMachine") -> bool:
    state_machine.reset_state_machine()
    return True



from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine

from robot_interface.models.exceptions.robot_exceptions import ErrorMessage
from robot_interface.models.mission.status import MissionStatus, TaskStatus


def finish_mission(state_machine: "StateMachine") -> bool:
    fail_statuses: List[TaskStatus] = [
        TaskStatus.Cancelled,
        TaskStatus.Failed,
    ]
    partially_fail_statuses = fail_statuses + [TaskStatus.PartiallySuccessful]

    if len(state_machine.current_mission.tasks) == 0:
        state_machine.current_mission.status = MissionStatus.Successful
    elif all(
        task.status in fail_statuses for task in state_machine.current_mission.tasks
    ):
        state_machine.current_mission.error_message = ErrorMessage(
            error_reason=None,
            error_description="The mission failed because all tasks in the mission "
            "failed",
        )
        state_machine.current_mission.status = MissionStatus.Failed
    elif any(
        task.status in partially_fail_statuses
        for task in state_machine.current_mission.tasks
    ):
        state_machine.current_mission.status = MissionStatus.PartiallySuccessful
    else:
        state_machine.current_mission.status = MissionStatus.Successful
    state_machine._finalize()

    state_machine.current_task = None
    state_machine.send_task_status()
    return True



from typing import TYPE_CHECKING, List

from isar.config.settings import settings
from isar.eventhandlers.eventhandler import EventHandlerBase, EventHandlerMapping
from isar.models.events import Event
from robot_interface.models.mission.status import RobotStatus

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Recharging(EventHandlerBase):

    def __init__(self, state_machine: "StateMachine"):
        shared_state = state_machine.shared_state

        def robot_battery_level_updated_handler(event: Event[float]):
            battery_level: float = event.check()
            if battery_level >= settings.ROBOT_BATTERY_RECHARGE_THRESHOLD:
                return state_machine.robot_recharged  # type: ignore
            return None

        def robot_offline_handler(event: Event[RobotStatus]):
            robot_status: RobotStatus = event.check()
            if robot_status == RobotStatus.Offline:
                return state_machine.robot_went_offline  # type: ignore

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping(
                name="robot_battery_update_event",
                event=shared_state.robot_battery_level,
                handler=robot_battery_level_updated_handler,
            ),
            EventHandlerMapping(
                name="robot_offline_event",
                event=shared_state.robot_status,
                handler=robot_offline_handler,
            ),
        ]
        super().__init__(
            state_name="recharging",
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )



from enum import Enum


class States(str, Enum):
    Monitor = "monitor"
    ReturningHome = "returning_home"
    Stopping = "stopping"
    Paused = "paused"
    AwaitNextMission = "await_next_mission"
    Home = "home"
    RobotStandingStill = "robot_standing_still"
    Offline = "offline"
    BlockedProtectiveStop = "blocked_protective_stop"
    UnknownStatus = "unknown_status"
    InterventionNeeded = "intervention_needed"
    Recharging = "recharging"

    def __repr__(self):
        return self.value



import logging
import time
from threading import Event, Thread
from typing import Optional

from isar.config.settings import settings
from isar.models.events import RobotServiceEvents
from isar.services.utilities.threaded_request import ThreadedRequest
from robot_interface.models.exceptions.robot_exceptions import (
    ErrorMessage,
    RobotCommunicationException,
    RobotCommunicationTimeoutException,
    RobotException,
    RobotTaskStatusException,
)
from robot_interface.models.mission.status import TaskStatus
from robot_interface.robot_interface import RobotInterface


class RobotTaskStatusThread(Thread):
    def __init__(
        self,
        robot_service_events: RobotServiceEvents,
        robot: RobotInterface,
        signal_thread_quitting: Event,
        task_id: str,
    ):
        self.logger = logging.getLogger("robot")
        self.robot_service_events: RobotServiceEvents = robot_service_events
        self.robot: RobotInterface = robot
        self.start_mission_thread: Optional[ThreadedRequest] = None
        self.signal_thread_quitting: Event = signal_thread_quitting
        self.task_id: str = task_id
        Thread.__init__(self, name="Robot task status thread")

    def run(self) -> None:
        task_status: TaskStatus = TaskStatus.NotStarted
        failed_task_error: Optional[ErrorMessage] = None
        request_status_failure_counter: int = 0

        while (
            request_status_failure_counter
            < settings.REQUEST_STATUS_FAILURE_COUNTER_LIMIT
        ):
            if self.signal_thread_quitting.wait(0):
                return
            if request_status_failure_counter > 0:
                time.sleep(settings.REQUEST_STATUS_COMMUNICATION_RECONNECT_DELAY)

            try:
                task_status = self.robot.task_status(self.task_id)
                request_status_failure_counter = 0
            except (
                RobotCommunicationTimeoutException,
                RobotCommunicationException,
            ) as e:
                request_status_failure_counter += 1
                self.logger.error(
                    f"Failed to get task status "
                    f"{request_status_failure_counter} times because: "
                    f"{e.error_description}"
                )

                failed_task_error = ErrorMessage(
                    error_reason=e.error_reason,
                    error_description=e.error_description,
                )
                continue

            except RobotTaskStatusException as e:
                failed_task_error = ErrorMessage(
                    error_reason=e.error_reason,
                    error_description=e.error_description,
                )
                break

            except RobotException as e:
                failed_task_error = ErrorMessage(
                    error_reason=e.error_reason,
                    error_description=e.error_description,
                )
                break

            self.robot_service_events.task_status_updated.trigger_event(task_status)
            return

        self.robot_service_events.task_status_failed.trigger_event(failed_task_error)



from collections.abc import Callable
from typing import TYPE_CHECKING, List, Optional

from isar.eventhandlers.eventhandler import EventHandlerBase, EventHandlerMapping
from isar.models.events import Event
from isar.state_machine.utils.common_event_handlers import return_home_event_handler

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class InterventionNeeded(EventHandlerBase):

    def __init__(self, state_machine: "StateMachine"):
        events = state_machine.events

        def release_intervention_needed_handler(
            event: Event[bool],
        ) -> Optional[Callable]:
            if event.consume_event():
                state_machine.events.api_requests.release_intervention_needed.response.trigger_event(
                    True
                )
                return state_machine.release_intervention_needed  # type: ignore
            return None

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping(
                name="return_home_event",
                event=events.api_requests.return_home.request,
                handler=lambda event: return_home_event_handler(state_machine, event),
            ),
            EventHandlerMapping(
                name="release_intervention_needed_event",
                event=events.api_requests.release_intervention_needed.request,
                handler=release_intervention_needed_handler,
            ),
        ]
        super().__init__(
            state_name="intervention_needed",
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )



import time
from collections import deque
from http import HTTPStatus
from threading import Thread
from typing import List, Optional, cast
from uuid import uuid4

import pytest
from fastapi import HTTPException
from pytest_mock import MockerFixture

from isar.config.settings import settings
from isar.eventhandlers.eventhandler import (
    EventHandlerBase,
    EventHandlerMapping,
    TimeoutHandlerMapping,
)
from isar.modules import ApplicationContainer
from isar.robot.robot import Robot
from isar.robot.robot_status import RobotStatusThread
from isar.services.utilities.scheduling_utilities import SchedulingUtilities
from isar.state_machine.state_machine import StateMachine, main
from isar.state_machine.states_enum import States
from isar.storage.storage_interface import StorageInterface
from isar.storage.uploader import Uploader
from robot_interface.models.exceptions.robot_exceptions import (
    ErrorReason,
    RobotException,
)
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import RobotStatus, TaskStatus
from robot_interface.models.mission.task import TakeImage, Task
from tests.test_double.pose import DummyPose
from tests.test_double.robot_interface import (
    StubRobot,
    StubRobotBlockedProtectiveStopToRobotStandingStillTest,
    StubRobotHomeToRobotStandingStillTest,
    StubRobotOfflineToHomeTest,
    StubRobotOfflineToRobotStandingStillTest,
)
from tests.test_double.task import StubTask


class StateMachineThreadMock(object):
    def __init__(self, container: ApplicationContainer) -> None:
        self.state_machine: StateMachine = container.state_machine()
        self._thread: Thread = Thread(target=main, args=[self.state_machine])

    def start(self):
        self._thread.start()

    def join(self):
        self.state_machine.terminate()
        self._thread.join()


class UploaderThreadMock(object):
    def __init__(self, container: ApplicationContainer) -> None:
        self.uploader: Uploader = container.uploader()
        self._thread: Thread = Thread(target=self.uploader.run)

    def start(self):
        self._thread.start()

    def join(self):
        self.uploader.stop()
        self._thread.join()


class RobotServiceThreadMock(object):
    def __init__(self, robot_service: Robot) -> None:
        self.robot_service: Robot = robot_service

    def start(self) -> None:
        self._thread: Thread = Thread(target=self.robot_service.run)
        self._thread.start()

    def join(self):
        self.robot_service.stop()
        self._thread.join()


def test_initial_unknown_status(state_machine) -> None:
    assert state_machine.state == "unknown_status"


def test_reset_state_machine(state_machine) -> None:
    state_machine.reset_state_machine()

    assert state_machine.current_task is None
    assert state_machine.current_mission is None


def test_state_machine_transitions_when_running_full_mission(
    container: ApplicationContainer,
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
    mocker,
) -> None:
    state_machine_thread.state_machine.await_next_mission_state.timers[
        0
    ].timeout_in_seconds = 0.01
    state_machine_thread.start()
    mocker.patch.object(StubRobot, "robot_status", return_value=RobotStatus.Home)
    robot_service_thread.start()
    task_1: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    task_2: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    mission: Mission = Mission(name="Dummy misson", tasks=[task_1, task_2])

    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()
    scheduling_utilities.start_mission(mission=mission)
    time.sleep(3)  # Allow enough time to run mission and return home

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.Home,
            States.Monitor,
            States.AwaitNextMission,
            States.ReturningHome,
            States.Home,
        ]
    )


def test_state_machine_battery_too_low_to_start_mission(
    container: ApplicationContainer,
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
    mocker,
) -> None:
    state_machine_thread.state_machine.await_next_mission_state.timers[
        0
    ].timeout_in_seconds = 0.01
    state_machine_thread.start()
    mocker.patch.object(StubRobot, "robot_status", return_value=RobotStatus.Home)
    mocker.patch.object(StubRobot, "get_battery_level", return_value=10.0)
    robot_service_thread.start()
    task_1: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    task_2: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    mission: Mission = Mission(name="Dummy misson", tasks=[task_1, task_2])

    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()

    with pytest.raises(HTTPException) as exception_details:
        scheduling_utilities.start_mission(mission=mission)
        assert exception_details.value.status_code == 408

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.Home,
        ]
    )


def test_return_home_not_cancelled_when_battery_is_low(
    container: ApplicationContainer,
) -> None:
    state_machine: StateMachine = container.state_machine()
    state_machine.shared_state.robot_battery_level.trigger_event(10.0)

    events = state_machine.events

    returning_home_state: EventHandlerBase = cast(
        EventHandlerBase, state_machine.returning_home_state
    )
    event_handler: Optional[EventHandlerMapping] = (
        returning_home_state.get_event_handler_by_name("start_mission_event")
    )

    assert event_handler is not None

    event_handler.event.trigger_event(True)
    transition = event_handler.handler(event_handler.event)

    assert transition is None
    assert events.api_requests.start_mission.response.has_event()
    start_mission_event_response = events.api_requests.start_mission.response.check()
    assert not start_mission_event_response.mission_started


def test_return_home_starts_when_battery_is_low(
    container: ApplicationContainer,
) -> None:
    state_machine: StateMachine = container.state_machine()
    state_machine.shared_state.robot_battery_level.trigger_event(10.0)

    await_next_mission_state: EventHandlerBase = cast(
        EventHandlerBase, state_machine.await_next_mission_state
    )
    timer: Optional[TimeoutHandlerMapping] = (
        await_next_mission_state.get_event_timer_by_name("should_return_home_timer")
    )

    assert timer is not None

    transition = timer.handler()

    assert transition is state_machine.request_return_home  # type: ignore


def test_monitor_goes_to_return_home_when_battery_low(
    container: ApplicationContainer,
) -> None:
    state_machine: StateMachine = container.state_machine()

    state_machine.mission_ongoing = True

    task_1: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    state_machine.current_mission = Mission(name="Dummy misson", tasks=[task_1])
    state_machine.current_task = task_1

    monitor_state: EventHandlerBase = cast(
        EventHandlerBase, state_machine.monitor_state
    )
    event_handler: Optional[EventHandlerMapping] = (
        monitor_state.get_event_handler_by_name("robot_battery_update_event")
    )

    assert event_handler is not None

    event_handler.event.trigger_event(10.0, timeout=1)
    transition = event_handler.handler(event_handler.event)

    assert transition is state_machine.request_return_home  # type: ignore
    assert not state_machine.events.mqtt_queue.empty()

    mqtt_message = state_machine.events.mqtt_queue.get(block=False)
    assert mqtt_message is not None
    mqtt_payload_topic = mqtt_message[0]
    assert mqtt_payload_topic is settings.TOPIC_ISAR_MISSION_ABORTED
    assert state_machine.events.mqtt_queue.get()[0]


def test_return_home_goes_to_recharging_when_battery_low(
    container: ApplicationContainer,
) -> None:
    state_machine: StateMachine = container.state_machine()
    state_machine.shared_state.robot_battery_level.trigger_event(10.0)

    state_machine.mission_ongoing = True
    state_machine.current_task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )

    returning_home_state: EventHandlerBase = cast(
        EventHandlerBase, state_machine.returning_home_state
    )
    event_handler: Optional[EventHandlerMapping] = (
        returning_home_state.get_event_handler_by_name("task_status_event")
    )

    assert event_handler is not None

    event_handler.event.trigger_event(TaskStatus.Successful)
    transition = event_handler.handler(event_handler.event)

    assert transition is state_machine.starting_recharging  # type: ignore


def test_recharging_goes_to_home_when_battery_high(
    container: ApplicationContainer,
) -> None:
    state_machine: StateMachine = container.state_machine()

    recharging_state: EventHandlerBase = cast(
        EventHandlerBase, state_machine.recharging_state
    )
    event_handler: Optional[EventHandlerMapping] = (
        recharging_state.get_event_handler_by_name("robot_battery_update_event")
    )

    assert event_handler is not None

    event_handler.event.trigger_event(99.9)
    transition = event_handler.handler(event_handler.event)

    assert transition is state_machine.robot_recharged  # type: ignore


def test_recharging_continues_when_battery_low(
    container: ApplicationContainer,
) -> None:
    state_machine: StateMachine = container.state_machine()

    recharging_state: EventHandlerBase = cast(
        EventHandlerBase, state_machine.recharging_state
    )
    event_handler: Optional[EventHandlerMapping] = (
        recharging_state.get_event_handler_by_name("robot_battery_update_event")
    )

    assert event_handler is not None

    event_handler.event.trigger_event(10.0)
    transition = event_handler.handler(event_handler.event)

    assert transition is None


def test_state_machine_failed_dependency(
    container: ApplicationContainer,
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
    mocker,
) -> None:
    state_machine_thread.state_machine.await_next_mission_state.timers[
        0
    ].timeout_in_seconds = 0.01

    mocker.patch.object(StubRobot, "task_status", return_value=TaskStatus.Failed)

    task_1: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    task_2: Task = TakeImage(
        target=DummyPose.default_pose().position, robot_pose=DummyPose.default_pose()
    )
    mission: Mission = Mission(name="Dummy misson", tasks=[task_1, task_2])

    mocker.patch.object(StubRobot, "task_status", return_value=TaskStatus.Failed)

    state_machine_thread.start()
    robot_service_thread.start()

    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()
    scheduling_utilities.start_mission(mission=mission)
    time.sleep(3)  # Allow the state machine to transition through the mission

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.RobotStandingStill,
            States.Monitor,
            States.AwaitNextMission,
            States.ReturningHome,
            States.ReturningHome,
            States.ReturningHome,
            States.ReturningHome,
            States.ReturningHome,
            States.InterventionNeeded,
        ]
    )


def test_state_machine_with_successful_collection(
    container: ApplicationContainer,
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
    uploader_thread: UploaderThreadMock,
    mocker,
) -> None:
    mocker.patch.object(StubRobot, "robot_status", return_value=RobotStatus.Home)

    storage_mock: StorageInterface = container.storage_handlers(List[StorageInterface])[
        0
    ]

    mocker.patch.object(
        RobotStatusThread, "_is_ready_to_poll_for_status", return_value=True
    )

    mission: Mission = Mission(name="Dummy misson", tasks=[StubTask.take_image()])
    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()

    state_machine_thread.state_machine.await_next_mission_state.timers[
        0
    ].timeout_in_seconds = 0.01
    state_machine_thread.start()
    uploader_thread.start()

    robot_service_thread.start()

    scheduling_utilities.start_mission(mission=mission)
    time.sleep(3)  # Allow enough time to run mission and return home

    expected_stored_items = 1
    assert len(storage_mock.stored_inspections) == expected_stored_items  # type: ignore
    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.Home,
            States.Monitor,
            States.AwaitNextMission,
            States.ReturningHome,
            States.Home,
        ]
    )


def test_state_machine_with_unsuccessful_collection(
    container: ApplicationContainer,
    mocker,
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
    uploader_thread: UploaderThreadMock,
) -> None:
    mocker.patch.object(StubRobot, "robot_status", return_value=RobotStatus.Home)

    storage_mock: StorageInterface = container.storage_handlers(List[StorageInterface])[
        0
    ]

    mocker.patch.object(StubRobot, "get_inspection", return_value=None)

    mocker.patch.object(
        RobotStatusThread, "_is_ready_to_poll_for_status", return_value=True
    )

    state_machine_thread.state_machine.await_next_mission_state.timers[
        0
    ].timeout_in_seconds = 0.01
    state_machine_thread.start()
    robot_service_thread.start()
    uploader_thread.start()

    mission: Mission = Mission(name="Dummy misson", tasks=[StubTask.take_image()])
    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()
    scheduling_utilities.start_mission(mission=mission)
    time.sleep(3)  # Allow enough time to run mission and return home

    expected_stored_items = 0
    assert len(storage_mock.stored_inspections) == expected_stored_items  # type: ignore

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.Home,
            States.Monitor,
            States.AwaitNextMission,
            States.ReturningHome,
            States.Home,
        ]
    )


def test_state_machine_with_successful_mission_stop(
    container: ApplicationContainer,
    robot_service_thread: RobotServiceThreadMock,
    state_machine_thread: StateMachineThreadMock,
    uploader_thread: UploaderThreadMock,
    mocker,
) -> None:
    mocker.patch.object(StubRobot, "robot_status", return_value=RobotStatus.Home)
    mocker.patch.object(StubRobot, "task_status", return_value=TaskStatus.InProgress)

    mocker.patch.object(
        RobotStatusThread, "_is_ready_to_poll_for_status", return_value=True
    )

    # Set the return home delay to a higher value than the test needs to run
    state_machine_thread.state_machine.await_next_mission_state.timers[
        0
    ].timeout_in_seconds = 100

    mission: Mission = Mission(
        name="Dummy misson", tasks=[StubTask.take_image() for _ in range(0, 20)]
    )

    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()

    state_machine_thread.start()
    robot_service_thread.start()
    uploader_thread.start()

    scheduling_utilities.start_mission(mission=mission)
    scheduling_utilities.stop_mission(mission_id="")
    time.sleep(1)  # Allow enough time to stop the mission

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.Home,
            States.Monitor,
            States.Stopping,
            States.AwaitNextMission,
        ]
    )


def test_state_machine_with_unsuccessful_mission_stop_with_mission_id(
    container: ApplicationContainer,
    mocker: MockerFixture,
    state_machine_thread: StateMachineThreadMock,
    caplog: pytest.LogCaptureFixture,
    robot_service_thread: RobotServiceThreadMock,
) -> None:
    mission: Mission = Mission(name="Dummy misson", tasks=[StubTask.take_image()])

    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()
    mocker.patch.object(StubRobot, "task_status", return_value=TaskStatus.InProgress)
    mocker.patch.object(
        StubRobot, "stop", side_effect=_mock_robot_exception_with_message
    )

    state_machine_thread.state_machine.sleep_time = 0

    state_machine_thread.start()
    robot_service_thread.start()

    scheduling_utilities.start_mission(mission=mission)
    time.sleep(1)
    with pytest.raises(HTTPException) as exception_details:
        scheduling_utilities.stop_mission(str(uuid4()))

    assert exception_details.value.status_code == HTTPStatus.NOT_FOUND.value
    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.RobotStandingStill,
            States.Monitor,
        ]
    )


def test_state_machine_with_unsuccessful_mission_stop(
    container: ApplicationContainer,
    mocker: MockerFixture,
    state_machine_thread: StateMachineThreadMock,
    caplog: pytest.LogCaptureFixture,
    robot_service_thread: RobotServiceThreadMock,
) -> None:
    mission: Mission = Mission(name="Dummy misson", tasks=[StubTask.take_image()])

    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()
    mocker.patch.object(StubRobot, "task_status", return_value=TaskStatus.InProgress)
    mocker.patch.object(
        StubRobot, "stop", side_effect=_mock_robot_exception_with_message
    )

    state_machine_thread.state_machine.sleep_time = 0

    state_machine_thread.start()
    robot_service_thread.start()

    scheduling_utilities.start_mission(mission=mission)
    time.sleep(1)
    with pytest.raises(HTTPException) as exception_details:
        scheduling_utilities.stop_mission()

    expected_log = (
        "Be aware that the robot may still be "
        "moving even though a stop has been attempted"
    )
    assert exception_details.value.status_code == HTTPStatus.CONFLICT.value
    assert expected_log in caplog.text
    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.RobotStandingStill,
            States.Monitor,
            States.Stopping,
            States.Monitor,
        ]
    )


def test_state_machine_with_unsuccessful_return_home_stop(
    container: ApplicationContainer,
    mocker: MockerFixture,
    state_machine_thread: StateMachineThreadMock,
    caplog: pytest.LogCaptureFixture,
    robot_service_thread: RobotServiceThreadMock,
) -> None:
    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()
    mocker.patch.object(StubRobot, "task_status", return_value=TaskStatus.InProgress)
    mocker.patch.object(
        StubRobot, "stop", side_effect=_mock_robot_exception_with_message
    )

    state_machine_thread.state_machine.sleep_time = 0

    state_machine_thread.start()
    robot_service_thread.start()

    scheduling_utilities.return_home()
    time.sleep(1)
    with pytest.raises(HTTPException) as exception_details:
        scheduling_utilities.stop_mission()

    expected_log = (
        "Be aware that the robot may still be "
        "moving even though a stop has been attempted"
    )
    assert exception_details.value.status_code == HTTPStatus.CONFLICT.value
    assert expected_log in caplog.text
    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.RobotStandingStill,
            States.ReturningHome,
            States.Stopping,
            States.ReturningHome,
        ]
    )


def test_state_machine_with_successful_return_home_stop(
    container: ApplicationContainer,
    mocker: MockerFixture,
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
) -> None:
    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()
    mocker.patch.object(StubRobot, "task_status", return_value=TaskStatus.InProgress)

    state_machine_thread.state_machine.sleep_time = 0

    state_machine_thread.start()
    robot_service_thread.start()

    scheduling_utilities.return_home()
    time.sleep(1)
    scheduling_utilities.stop_mission()

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.RobotStandingStill,
            States.ReturningHome,
            States.Stopping,
            States.RobotStandingStill,
        ]
    )


def test_state_machine_with_mission_start_during_return_home_without_queueing_stop_response(
    container: ApplicationContainer,
    mocker: MockerFixture,
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
) -> None:
    mocker.patch.object(StubRobot, "robot_status", return_value=RobotStatus.Available)
    mission: Mission = Mission(name="Dummy misson", tasks=[StubTask.take_image()])
    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()
    mocker.patch.object(StubRobot, "task_status", return_value=TaskStatus.InProgress)

    state_machine_thread.state_machine.sleep_time = 0

    state_machine_thread.start()
    robot_service_thread.start()

    scheduling_utilities.return_home()
    time.sleep(1)
    scheduling_utilities.start_mission(mission=mission)

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.RobotStandingStill,
            States.ReturningHome,
            States.Stopping,
            States.RobotStandingStill,
            States.Monitor,
        ]
    )
    assert (
        not state_machine_thread.state_machine.events.api_requests.start_mission.request.has_event()
    )


def test_state_machine_with_return_home_failure_successful_retries(
    container: ApplicationContainer,
    mocker: MockerFixture,
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
) -> None:
    state_machine_thread.state_machine.await_next_mission_state.timers[
        0
    ].timeout_in_seconds = 0.01
    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()
    state_machine_thread.start()
    mocker.patch.object(StubRobot, "robot_status", return_value=RobotStatus.Available)
    mocker.patch.object(
        StubRobot, "task_status", side_effect=[TaskStatus.Failed, TaskStatus.Successful]
    )
    robot_service_thread.start()

    scheduling_utilities.return_home()
    time.sleep(3)  # Allow enough time to run mission and return home

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.RobotStandingStill,
            States.ReturningHome,
            States.ReturningHome,
            States.Home,
            States.RobotStandingStill,
        ]
    )


def test_state_machine_with_return_home_failure_unsuccessful_retries_twice(
    container: ApplicationContainer,
    mocker: MockerFixture,
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
) -> None:
    state_machine_thread.state_machine.await_next_mission_state.timers[
        0
    ].timeout_in_seconds = 0.01
    scheduling_utilities: SchedulingUtilities = container.scheduling_utilities()
    state_machine_thread.start()
    mocker.patch.object(StubRobot, "robot_status", return_value=RobotStatus.Available)
    mocker.patch.object(StubRobot, "task_status", return_value=TaskStatus.Failed)
    robot_service_thread.start()

    scheduling_utilities.return_home()
    time.sleep(3)  # Allow enough time to run mission and return home
    scheduling_utilities.release_intervention_needed()
    scheduling_utilities.return_home()
    time.sleep(3)  # Allow enough time to run mission and return home

    assert state_machine_thread.state_machine.transitions_list == deque(
        [
            States.UnknownStatus,
            States.RobotStandingStill,
            States.ReturningHome,
            States.ReturningHome,
            States.ReturningHome,
            States.ReturningHome,
            States.ReturningHome,
            States.InterventionNeeded,
            States.UnknownStatus,
            States.RobotStandingStill,
            States.ReturningHome,
            States.ReturningHome,
            States.ReturningHome,
            States.ReturningHome,
            States.ReturningHome,
            States.InterventionNeeded,
        ]
    )


def test_state_machine_offline_to_robot_standing_still(
    state_machine_thread, robot_service_thread, mocker
) -> None:
    # Robot status check happens every 5 seconds by default, so we mock the behavior
    # to poll for status imediately
    mocker.patch.object(
        RobotStatusThread, "_is_ready_to_poll_for_status", return_value=True
    )

    robot_service_thread.robot_service.robot = StubRobotOfflineToRobotStandingStillTest(
        robot_service_thread.robot_service.shared_state.state
    )
    state_machine_thread.start()
    robot_service_thread.start()
    time.sleep(1)

    assert state_machine_thread.state_machine.transitions_list == deque(
        [States.UnknownStatus, States.Offline, States.RobotStandingStill]
    )


def test_state_machine_idle_to_blocked_protective_stop_to_idle(
    state_machine_thread: StateMachineThreadMock,
    robot_service_thread: RobotServiceThreadMock,
    mocker,
) -> None:
    # Robot status check happens every 5 seconds by default, so we mock the behavior
    # to poll for status imediately
    mocker.patch.object(
        RobotStatusThread, "_is_ready_to_poll_for_status", return_value=True
    )

    robot_service_thread.robot_service.robot = (
        StubRobotBlockedProtectiveStopToRobotStandingStillTest(
            robot_service_thread.robot_service.shared_state.state
        )
    )

    state_machine_thread.start()
    robot_service_thread.start()
    time.sleep(1)

    assert state_machine_thread.state_machine.transitions_list == deque(
        [States.UnknownStatus, States.BlockedProtectiveStop, States.RobotStandingStill]
    )


def test_state_machine_home_to_robot_standing_still(
    mocker, state_machine_thread, robot_service_thread
) -> None:
    # Robot status check happens every 5 seconds by default, so we mock the behavior
    # to poll for status imediately
    mocker.patch.object(
        RobotStatusThread, "_is_ready_to_poll_for_status", return_value=True
    )
    robot_service_thread.robot_service.robot = StubRobotHomeToRobotStandingStillTest(
        robot_service_thread.robot_service.shared_state.state
    )

    state_machine_thread.start()
    robot_service_thread.start()
    time.sleep(1)

    assert state_machine_thread.state_machine.transitions_list == deque(
        [States.UnknownStatus, States.Home, States.RobotStandingStill]
    )


def test_state_machine_offline_to_home(
    mocker, state_machine_thread, robot_service_thread
) -> None:
    # Robot status check happens every 5 seconds by default, so we mock the behavior
    # to poll for status imediately
    mocker.patch.object(
        RobotStatusThread, "_is_ready_to_poll_for_status", return_value=True
    )
    robot_service_thread.robot_service.robot = StubRobotOfflineToHomeTest(
        robot_service_thread.robot_service.shared_state.state
    )

    state_machine_thread.start()
    robot_service_thread.start()
    time.sleep(1)

    assert state_machine_thread.state_machine.transitions_list == deque(
        [States.UnknownStatus, States.Offline, States.Home]
    )


def _mock_robot_exception_with_message() -> RobotException:
    raise RobotException(
        error_reason=ErrorReason.RobotUnknownErrorException,
        error_description="This is an example error description",
    )



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine

from isar.apis.models.models import ControlMissionResponse
from robot_interface.models.mission.status import MissionStatus, TaskStatus


def trigger_stop_mission_event(state_machine: "StateMachine") -> bool:
    state_machine.events.state_machine_events.stop_mission.trigger_event(True)
    return True


def stop_mission_cleanup(state_machine: "StateMachine") -> bool:
    if state_machine.current_mission is None:
        state_machine._queue_empty_response()
        state_machine.reset_state_machine()
        return True

    state_machine.current_mission.status = MissionStatus.Cancelled

    for task in state_machine.current_mission.tasks:
        if task.status in [
            TaskStatus.NotStarted,
            TaskStatus.InProgress,
            TaskStatus.Paused,
        ]:
            task.status = TaskStatus.Cancelled

    stopped_mission_response: ControlMissionResponse = (
        state_machine._make_control_mission_response()
    )
    state_machine.events.api_requests.stop_mission.response.put(
        stopped_mission_response
    )
    state_machine.publish_task_status(task=state_machine.current_task)
    state_machine._finalize()
    return True


def stop_mission_failed(state_machine: "StateMachine") -> bool:
    stopped_mission_response: ControlMissionResponse = (
        state_machine._make_control_mission_response()
    )
    state_machine.events.api_requests.stop_mission.response.put(
        stopped_mission_response
    )
    return True


def stop_return_home_mission_cleanup(state_machine: "StateMachine") -> bool:
    if state_machine.current_mission is None:
        state_machine._queue_empty_response()
        state_machine.reset_state_machine()
        return True

    if not state_machine.events.api_requests.start_mission.request.has_event():
        state_machine.current_mission.status = MissionStatus.Cancelled

        for task in state_machine.current_mission.tasks:
            if task.status in [
                TaskStatus.NotStarted,
                TaskStatus.InProgress,
                TaskStatus.Paused,
            ]:
                task.status = TaskStatus.Cancelled

        stopped_mission_response: ControlMissionResponse = (
            state_machine._make_control_mission_response()
        )
        state_machine.events.api_requests.stop_mission.response.put(
            stopped_mission_response
        )

    state_machine._finalize()
    return True


def stop_return_home_mission_failed(state_machine: "StateMachine") -> bool:
    if state_machine.events.api_requests.start_mission.request.has_event():
        return True
    stopped_mission_response: ControlMissionResponse = (
        state_machine._make_control_mission_response()
    )
    state_machine.events.api_requests.stop_mission.response.put(
        stopped_mission_response
    )
    return True



from typing import TYPE_CHECKING, List

from isar.state_machine.transitions.functions.robot_status import (
    is_available,
    is_blocked_protective_stop,
    is_home,
    is_offline,
)
from isar.state_machine.transitions.functions.utils import def_transition

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


def get_robot_status_transitions(state_machine: "StateMachine") -> List[dict]:
    robot_status_transitions: List[dict] = [
        {
            "trigger": "robot_status_changed",
            "source": [
                state_machine.home_state,
                state_machine.blocked_protective_stopping_state,
                state_machine.offline_state,
                state_machine.unknown_status_state,
            ],
            "dest": state_machine.robot_standing_still_state,
            "conditions": def_transition(state_machine, is_available),
        },
        {
            "trigger": "robot_status_changed",
            "source": [
                state_machine.robot_standing_still_state,
                state_machine.blocked_protective_stopping_state,
                state_machine.offline_state,
                state_machine.unknown_status_state,
            ],
            "dest": state_machine.home_state,
            "conditions": def_transition(state_machine, is_home),
        },
        {
            "trigger": "robot_status_changed",
            "source": [
                state_machine.robot_standing_still_state,
                state_machine.home_state,
                state_machine.offline_state,
                state_machine.unknown_status_state,
            ],
            "dest": state_machine.blocked_protective_stopping_state,
            "conditions": def_transition(state_machine, is_blocked_protective_stop),
        },
        {
            "trigger": "robot_status_changed",
            "source": [
                state_machine.robot_standing_still_state,
                state_machine.home_state,
                state_machine.blocked_protective_stopping_state,
                state_machine.unknown_status_state,
            ],
            "dest": state_machine.offline_state,
            "conditions": def_transition(state_machine, is_offline),
        },
        {
            "trigger": "robot_status_changed",
            "source": [
                state_machine.robot_standing_still_state,
                state_machine.home_state,
                state_machine.blocked_protective_stopping_state,
                state_machine.offline_state,
                state_machine.unknown_status_state,
            ],
            "dest": state_machine.unknown_status_state,
        },
        {
            "trigger": "robot_went_offline",
            "source": [state_machine.recharging_state],
            "dest": state_machine.offline_state,
        },
        {
            "trigger": "robot_recharged",
            "source": [state_machine.recharging_state],
            "dest": state_machine.home_state,
        },
    ]
    return robot_status_transitions



from typing import TYPE_CHECKING, List

from isar.eventhandlers.eventhandler import EventHandlerBase, EventHandlerMapping

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Paused(EventHandlerBase):

    def __init__(self, state_machine: "StateMachine"):
        events = state_machine.events

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping(
                name="stop_mission_event",
                event=events.api_requests.stop_mission.request,
                handler=lambda event: state_machine.stop if event.consume_event() else None,  # type: ignore
            ),
            EventHandlerMapping(
                name="resume_mission_event",
                event=events.api_requests.resume_mission.request,
                handler=lambda event: state_machine.resume if event.consume_event() else None,  # type: ignore
            ),
        ]
        super().__init__(
            state_name="paused",
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )



import logging
from typing import TYPE_CHECKING, Callable, List, Optional

from isar.eventhandlers.eventhandler import EventHandlerBase, EventHandlerMapping
from isar.models.events import Event
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Stopping(EventHandlerBase):

    def __init__(self, state_machine: "StateMachine"):
        logger = logging.getLogger("state_machine")
        events = state_machine.events

        def _failed_stop_event_handler(
            event: Event[ErrorMessage],
        ) -> Optional[Callable]:
            error_message: Optional[ErrorMessage] = event.consume_event()
            if error_message is not None:
                logger.warning(error_message.error_description)
                if (
                    state_machine.current_mission is not None
                    and state_machine.current_mission._is_return_to_home_mission()
                ):
                    return state_machine.return_home_mission_stopping_failed  # type: ignore
                else:
                    return state_machine.mission_stopping_failed  # type: ignore
            return None

        def _successful_stop_event_handler(event: Event[bool]) -> Optional[Callable]:
            if event.consume_event():
                if (
                    state_machine.current_mission is not None
                    and state_machine.current_mission._is_return_to_home_mission()
                ):
                    return state_machine.return_home_mission_stopped  # type: ignore
                else:
                    return state_machine.mission_stopped  # type: ignore
            return None

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping(
                name="failed_stop_event",
                event=events.robot_service_events.mission_failed_to_stop,
                handler=_failed_stop_event_handler,
            ),
            EventHandlerMapping(
                name="successful_stop_event",
                event=events.robot_service_events.mission_successfully_stopped,
                handler=_successful_stop_event_handler,
            ),
        ]
        super().__init__(
            state_name="stopping",
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )



from typing import TYPE_CHECKING, List

from isar.state_machine.transitions.functions.fail_mission import (
    report_failed_mission_and_finalize,
    report_failed_return_home_and_intervention_needed,
)
from isar.state_machine.transitions.functions.return_home import (
    reset_return_home_failure_counter,
    return_home_finished,
    set_return_home_status,
    should_retry_return_home,
    start_return_home_mission,
)
from isar.state_machine.transitions.functions.start_mission import (
    initialize_robot,
    trigger_start_mission_event,
)
from isar.state_machine.transitions.functions.utils import def_transition

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


def get_return_home_transitions(state_machine: "StateMachine") -> List[dict]:
    return_home_transitions: List[dict] = [
        {
            "trigger": "request_return_home",
            "source": [
                state_machine.await_next_mission_state,
                state_machine.home_state,
                state_machine.robot_standing_still_state,
                state_machine.intervention_needed_state,
                state_machine.monitor_state,
            ],
            "dest": state_machine.returning_home_state,
            "conditions": [
                def_transition(state_machine, start_return_home_mission),
                def_transition(state_machine, set_return_home_status),
                def_transition(state_machine, initialize_robot),
                def_transition(state_machine, initialize_robot),
            ],
            "before": def_transition(state_machine, trigger_start_mission_event),
        },
        {
            "trigger": "request_return_home",
            "source": state_machine.await_next_mission_state,
            "dest": state_machine.await_next_mission_state,
            "before": def_transition(state_machine, report_failed_mission_and_finalize),
        },
        {
            "trigger": "request_return_home",
            "source": state_machine.home_state,
            "dest": state_machine.home_state,
            "before": def_transition(state_machine, report_failed_mission_and_finalize),
        },
        {
            "trigger": "request_return_home",
            "source": state_machine.robot_standing_still_state,
            "dest": state_machine.robot_standing_still_state,
            "before": def_transition(state_machine, report_failed_mission_and_finalize),
        },
        {
            "trigger": "returned_home",
            "source": state_machine.returning_home_state,
            "dest": state_machine.home_state,
            "before": [
                def_transition(state_machine, reset_return_home_failure_counter),
                def_transition(state_machine, return_home_finished),
            ],
        },
        {
            "trigger": "starting_recharging",
            "source": state_machine.returning_home_state,
            "dest": state_machine.recharging_state,
            "before": [
                def_transition(state_machine, reset_return_home_failure_counter),
                def_transition(state_machine, return_home_finished),
            ],
        },
        {
            "trigger": "return_home_failed",
            "source": state_machine.returning_home_state,
            "dest": state_machine.returning_home_state,
            "conditions": [
                def_transition(state_machine, should_retry_return_home),
            ],
            "before": [
                def_transition(state_machine, report_failed_mission_and_finalize),
                def_transition(state_machine, start_return_home_mission),
                def_transition(state_machine, set_return_home_status),
                def_transition(state_machine, initialize_robot),
                def_transition(state_machine, trigger_start_mission_event),
            ],
        },
        {
            "trigger": "return_home_failed",
            "source": state_machine.returning_home_state,
            "dest": state_machine.intervention_needed_state,
            "before": [
                def_transition(
                    state_machine, report_failed_return_home_and_intervention_needed
                ),
                def_transition(state_machine, reset_return_home_failure_counter),
                def_transition(state_machine, report_failed_mission_and_finalize),
            ],
        },
        {
            "trigger": "release_intervention_needed",
            "source": state_machine.intervention_needed_state,
            "dest": state_machine.unknown_status_state,
        },
    ]
    return return_home_transitions



import logging
from copy import deepcopy
from typing import TYPE_CHECKING, Callable, List, Optional

from isar.config.settings import settings
from isar.eventhandlers.eventhandler import EventHandlerBase, EventHandlerMapping
from isar.models.events import Event
from isar.services.utilities.threaded_request import ThreadedRequest
from isar.state_machine.utils.common_event_handlers import (
    mission_failed_event_handler,
    mission_started_event_handler,
    stop_mission_event_handler,
    task_status_event_handler,
    task_status_failed_event_handler,
)
from robot_interface.models.mission.status import TaskStatus

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class Monitor(EventHandlerBase):

    def __init__(self, state_machine: "StateMachine"):
        logger = logging.getLogger("state_machine")
        events = state_machine.events
        shared_state = state_machine.shared_state

        def _pause_mission_event_handler(event: Event[bool]) -> Optional[Callable]:
            if event.consume_event():
                return state_machine.pause  # type: ignore
            return None

        def _handle_task_completed(task_status: TaskStatus):
            if state_machine.should_upload_inspections():
                get_inspection_thread = ThreadedRequest(
                    state_machine.queue_inspections_for_upload
                )
                get_inspection_thread.start_thread(
                    deepcopy(state_machine.current_mission),
                    deepcopy(state_machine.current_task),
                    logger,
                    name="State Machine Get Inspections",
                )

            state_machine.iterate_current_task()
            if state_machine.current_task is None:
                return state_machine.mission_finished  # type: ignore
            return None

        def _robot_battery_level_updated_handler(
            event: Event[float],
        ) -> Optional[Callable]:
            battery_level: float = event.check()
            if battery_level < settings.ROBOT_MISSION_BATTERY_START_THRESHOLD:
                state_machine.publish_mission_aborted(
                    "Robot battery too low to continue mission", True
                )
                state_machine._finalize()
                state_machine.logger.warning(
                    "Cancelling current mission due to low battery"
                )
                state_machine.current_mission = None
                state_machine.current_task = None
                return state_machine.request_return_home  # type: ignore
            return None

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping(
                name="stop_mission_event",
                event=events.api_requests.stop_mission.request,
                handler=lambda event: stop_mission_event_handler(state_machine, event),
            ),
            EventHandlerMapping(
                name="pause_mission_event",
                event=events.api_requests.pause_mission.request,
                handler=_pause_mission_event_handler,
            ),
            EventHandlerMapping(
                name="mission_started_event",
                event=events.robot_service_events.mission_started,
                handler=lambda event: mission_started_event_handler(
                    state_machine, event
                ),
            ),
            EventHandlerMapping(
                name="mission_failed_event",
                event=events.robot_service_events.mission_failed,
                handler=lambda event: mission_failed_event_handler(
                    state_machine, event
                ),
            ),
            EventHandlerMapping(
                name="task_status_failed_event",
                event=events.robot_service_events.task_status_failed,
                handler=lambda event: task_status_failed_event_handler(
                    state_machine, _handle_task_completed, event
                ),
            ),
            EventHandlerMapping(
                name="task_status_event",
                event=events.robot_service_events.task_status_updated,
                handler=lambda event: task_status_event_handler(
                    state_machine, _handle_task_completed, event
                ),
            ),
            EventHandlerMapping(
                name="robot_battery_update_event",
                event=shared_state.robot_battery_level,
                handler=_robot_battery_level_updated_handler,
            ),
        ]
        super().__init__(
            state_name="monitor",
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )



from typing import TYPE_CHECKING, Callable, List, Optional

from isar.eventhandlers.eventhandler import EventHandlerBase, EventHandlerMapping
from isar.models.events import Event
from isar.state_machine.utils.common_event_handlers import stop_mission_event_handler
from robot_interface.models.mission.status import RobotStatus

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class UnknownStatus(EventHandlerBase):

    def __init__(self, state_machine: "StateMachine"):
        events = state_machine.events
        shared_state = state_machine.shared_state

        def _robot_status_event_handler(
            event: Event[RobotStatus],
        ) -> Optional[Callable]:
            robot_status: RobotStatus = event.check()
            if (
                robot_status == RobotStatus.Home
                or robot_status == RobotStatus.Offline
                or robot_status == RobotStatus.BlockedProtectiveStop
                or robot_status == RobotStatus.Available
            ):
                return state_machine.robot_status_changed  # type: ignore
            return None

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping(
                name="stop_mission_event",
                event=events.api_requests.stop_mission.request,
                handler=lambda event: stop_mission_event_handler(state_machine, event),
            ),
            EventHandlerMapping(
                name="robot_status_event",
                event=shared_state.robot_status,
                handler=_robot_status_event_handler,
            ),
        ]
        super().__init__(
            state_name="unknown_status",
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )



from typing import TYPE_CHECKING, Callable, List, Optional

from isar.apis.models.models import MissionStartResponse
from isar.eventhandlers.eventhandler import EventHandlerBase, EventHandlerMapping
from isar.models.events import Event, EventTimeoutError
from isar.state_machine.utils.common_event_handlers import (
    mission_failed_event_handler,
    mission_started_event_handler,
    stop_mission_event_handler,
    task_status_event_handler,
    task_status_failed_event_handler,
)
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage, ErrorReason
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import TaskStatus

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class ReturningHome(EventHandlerBase):

    def __init__(self, state_machine: "StateMachine"):
        self.failed_return_home_attemps: int = 0
        events = state_machine.events

        def _handle_task_completed(status: TaskStatus):
            if status != TaskStatus.Successful:
                state_machine.current_mission.error_message = ErrorMessage(
                    error_reason=ErrorReason.RobotActionException,
                    error_description="Return home failed.",
                )
                self.failed_return_home_attemps += 1
                return state_machine.return_home_failed  # type: ignore

            if not state_machine.battery_level_is_above_mission_start_threshold():
                return state_machine.starting_recharging  # type: ignore
            else:
                return state_machine.returned_home  # type: ignore

        def _start_mission_event_handler(
            event: Event[Mission],
        ) -> Optional[Callable]:
            if event.has_event():
                if not state_machine.battery_level_is_above_mission_start_threshold():
                    try:
                        state_machine.events.api_requests.start_mission.response.trigger_event(
                            MissionStartResponse(
                                mission_id=None,
                                mission_started=False,
                                mission_not_started_reason="Robot battery too low",
                            ),
                            timeout=1,  # This conflict can happen if two API requests are received at the same time
                        )
                    except EventTimeoutError:
                        pass
                    return None
                return state_machine.stop  # type: ignore
            return None

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping(
                name="stop_mission_event",
                event=events.api_requests.stop_mission.request,
                handler=lambda event: stop_mission_event_handler(state_machine, event),
            ),
            EventHandlerMapping(
                name="mission_started_event",
                event=events.robot_service_events.mission_started,
                handler=lambda event: mission_started_event_handler(
                    state_machine, event
                ),
            ),
            EventHandlerMapping(
                name="mission_failed_event",
                event=events.robot_service_events.mission_failed,
                handler=lambda event: mission_failed_event_handler(
                    state_machine, event
                ),
            ),
            EventHandlerMapping(
                name="start_mission_event",
                event=events.api_requests.start_mission.request,
                handler=_start_mission_event_handler,
            ),
            EventHandlerMapping(
                name="task_status_failed_event",
                event=events.robot_service_events.task_status_failed,
                handler=lambda event: task_status_failed_event_handler(
                    state_machine, _handle_task_completed, event
                ),
            ),
            EventHandlerMapping(
                name="task_status_event",
                event=events.robot_service_events.task_status_updated,
                handler=lambda event: task_status_event_handler(
                    state_machine, _handle_task_completed, event
                ),
            ),
        ]
        super().__init__(
            state_name="returning_home",
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
        )



from typing import TYPE_CHECKING, List

from isar.config.settings import settings
from isar.eventhandlers.eventhandler import (
    EventHandlerBase,
    EventHandlerMapping,
    TimeoutHandlerMapping,
)
from isar.state_machine.utils.common_event_handlers import (
    return_home_event_handler,
    start_mission_event_handler,
    stop_mission_event_handler,
)

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class AwaitNextMission(EventHandlerBase):

    def __init__(self, state_machine: "StateMachine"):
        events = state_machine.events

        event_handlers: List[EventHandlerMapping] = [
            EventHandlerMapping(
                name="start_mission_event",
                event=events.api_requests.start_mission.request,
                handler=lambda event: start_mission_event_handler(
                    state_machine, event, events.api_requests.start_mission.response
                ),
            ),
            EventHandlerMapping(
                name="return_home_event",
                event=events.api_requests.return_home.request,
                handler=lambda event: return_home_event_handler(state_machine, event),
            ),
            EventHandlerMapping(
                name="stop_mission_event",
                event=events.api_requests.return_home.request,
                handler=lambda event: stop_mission_event_handler(state_machine, event),
            ),
        ]

        timers: List[TimeoutHandlerMapping] = [
            TimeoutHandlerMapping(
                name="should_return_home_timer",
                timeout_in_seconds=settings.RETURN_HOME_DELAY,
                handler=lambda: state_machine.request_return_home,  # type: ignore
            )
        ]

        super().__init__(
            state_name="await_next_mission",
            state_machine=state_machine,
            event_handler_mappings=event_handlers,
            timers=timers,
        )



from typing import List

from isar.storage.storage_interface import StorageException, StorageInterface
from robot_interface.models.inspection.inspection import Inspection
from robot_interface.models.mission.mission import Mission


class StorageFake(StorageInterface):
    will_fail: bool = False

    def __init__(self) -> None:
        self.stored_inspections: List[Inspection] = []

    def store(self, inspection: Inspection, mission: Mission):
        if self.will_fail:
            raise StorageException("Fake failed on purpose")
        self.stored_inspections.append(inspection)

    def blob_exists(self, inspection: Inspection) -> bool:
        return inspection in self.stored_inspections



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine

import time

from isar.apis.models.models import ControlMissionResponse
from isar.config.settings import settings
from robot_interface.models.exceptions.robot_exceptions import (
    RobotActionException,
    RobotException,
)
from robot_interface.models.mission.status import MissionStatus, TaskStatus


def pause_mission(state_machine: "StateMachine") -> bool:
    state_machine.logger.info("Pausing mission: %s", state_machine.current_mission.id)

    max_retries = settings.STATE_TRANSITION_NUM_RETIRES
    retry_interval = settings.STATE_TRANSITION_RETRY_INTERVAL_SEC

    for attempt in range(max_retries):
        try:
            state_machine.robot.pause()
            state_machine.current_mission.status = MissionStatus.Paused
            state_machine.current_task.status = TaskStatus.Paused

            paused_mission_response: ControlMissionResponse = (
                state_machine._make_control_mission_response()
            )
            state_machine.events.api_requests.pause_mission.response.put(
                paused_mission_response
            )

            state_machine.publish_mission_status()
            state_machine.publish_task_status(task=state_machine.current_task)

            state_machine.logger.info("Mission paused successfully.")
            return True
        except RobotActionException as e:
            state_machine.logger.warning(
                f"Attempt {attempt + 1} to pause mission failed: {e.error_description}"
            )
            time.sleep(retry_interval)
        except RobotException as e:
            state_machine.logger.warning(
                f"Attempt {attempt + 1} to pause mission raised a RobotException: {e.error_description}"
            )
            time.sleep(retry_interval)

    state_machine.logger.error("Failed to pause mission after multiple attempts.")
    return False



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine

from robot_interface.models.mission.status import MissionStatus, TaskStatus


def report_failed_mission_and_finalize(state_machine: "StateMachine") -> None:
    state_machine.current_task.status = TaskStatus.Failed
    state_machine.current_mission.status = MissionStatus.Failed
    state_machine.publish_task_status(task=state_machine.current_task)
    state_machine._finalize()


def report_failed_return_home_and_intervention_needed(
    state_machine: "StateMachine",
) -> None:
    error_message: str = "Return home failed."
    state_machine.publish_intervention_needed(error_message=error_message)



import logging
from copy import deepcopy
from http import HTTPStatus
from typing import List, TypeVar

from fastapi import HTTPException
from requests import HTTPError

from isar.apis.models.models import ControlMissionResponse
from isar.config.settings import settings
from isar.mission_planner.mission_planner_interface import (
    MissionNotFoundError,
    MissionPlannerError,
    MissionPlannerInterface,
)
from isar.models.events import (
    APIEvent,
    APIRequests,
    EventConflictError,
    Events,
    EventTimeoutError,
    SharedState,
)
from isar.state_machine.states_enum import States
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import MissionStatus

T1 = TypeVar("T1")
T2 = TypeVar("T2")


class SchedulingUtilities:
    """
    Contains utility functions for scheduling missions from the API. The class handles
    required thread communication through queues to the state machine.
    """

    def __init__(
        self,
        events: Events,
        shared_state: SharedState,
        mission_planner: MissionPlannerInterface,
        queue_timeout: int = settings.QUEUE_TIMEOUT,
    ):
        self.api_events: APIRequests = events.api_requests
        self.shared_state: SharedState = shared_state
        self.mission_planner: MissionPlannerInterface = mission_planner
        self.queue_timeout: int = queue_timeout
        self.logger = logging.getLogger("api")

    def get_state(self) -> States:
        """Return the current state of the state machine

        Raises
        ------
        HTTPException 500 Internal Server Error
            If the current state is not available on the queue
        """
        current_state = self.shared_state.state.check()
        if current_state is None:
            error_message: str = (
                "Internal Server Error - Current state of the state machine is unknown"
            )
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )
        return current_state

    def get_mission(self, mission_id: str) -> Mission:
        """Get the mission with mission_id from the current mission planner

        Raises
        ------
        HTTPException 404 Not Found
            If requested mission with mission_id is not found
        HTTPException 500 Internal Server Error
            If for some reason the mission can not be returned
        """
        try:
            return self.mission_planner.get_mission(mission_id)
        except HTTPError as e:
            self.logger.error(e)
            raise HTTPException(status_code=e.response.status_code)
        except MissionNotFoundError as e:
            self.logger.error(e)
            raise HTTPException(
                status_code=HTTPStatus.NOT_FOUND,
                detail=f"Mission with id '{mission_id}' not found",
            )
        except MissionPlannerError as e:
            self.logger.error(e)
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
                detail="Could not plan mission",
            )

    def verify_robot_capable_of_mission(
        self, mission: Mission, robot_capabilities: List[str]
    ) -> bool:
        """Verify that the robot is capable of performing this mission

        Raises
        ------
        HTTPException 400 Bad request
            If the robot is not capable of performing mission
        """
        missing_capabilities = {
            task.type for task in mission.tasks if task.type not in robot_capabilities
        }

        if missing_capabilities:
            error_message = (
                f"Bad Request - Robot is not capable of performing mission."
                f" Missing functionalities: {missing_capabilities}."
            )
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.BAD_REQUEST,
                detail=error_message,
            )

        return True

    def verify_state_machine_ready_to_receive_mission(self, state: States) -> bool:
        """Verify that the state machine is ready to receive a mission

        Raises
        ------
        HTTPException 409 Conflict
            If state machine is not home, robot standing still, awaiting next mission
            or returning home and therefore cannot start a new mission
        """
        if (
            state == States.RobotStandingStill
            or state == States.Home
            or state == States.AwaitNextMission
            or state == States.ReturningHome
        ):
            return True

        error_message = f"Conflict - Robot is not home, robot standing still, awaiting next mission or returning home - State: {state}"
        self.logger.warning(error_message)
        raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)

    def verify_state_machine_ready_to_receive_return_home_mission(
        self, state: States
    ) -> bool:
        """Verify that the state machine is ready to receive a return home mission

        Raises
        ------
        HTTPException 409 Conflict
            If state machine is not home, robot standing still or awaiting next mission
            and therefore cannot start a new return home mission
        """
        if (
            state == States.RobotStandingStill
            or state == States.Home
            or state == States.AwaitNextMission
        ):
            return True

        error_message = f"Conflict - Robot is not home, robot standing still or awaiting next mission - State: {state}"
        self.logger.warning(error_message)
        raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)

    def start_mission(
        self,
        mission: Mission,
    ) -> None:
        """Start mission

        Raises
        ------
        HTTTPException 408 Request timeout
            If there is a timeout while communicating with the state machine
        """
        try:
            mission_start_response = self._send_command(
                deepcopy(mission),
                self.api_events.start_mission,
            )
            if not mission_start_response.mission_started:
                self.logger.warning(
                    f"Mission failed to start - {mission_start_response.mission_not_started_reason}"
                )
                raise HTTPException(
                    status_code=HTTPStatus.CONFLICT,
                    detail=mission_start_response.mission_not_started_reason,
                )
        except EventConflictError:
            error_message = "Previous mission request is still being processed"
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except EventTimeoutError:
            error_message = (
                "State machine has entered a state which cannot start a mission"
            )
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        self.logger.info("OK - Mission started in ISAR")

    def return_home(
        self,
    ) -> None:
        """Start return home mission

        Raises
        ------
        HTTTPException 408 Request timeout
            If there is a timeout while communicating with the state machine
        """
        try:
            self._send_command(
                True,
                self.api_events.return_home,
            )
        except EventConflictError:
            error_message = "Previous return home request is still being processed"
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except EventTimeoutError:
            error_message = "State machine has entered a state which cannot start a return home mission"
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        self.logger.info("OK - Return home mission started in ISAR")

    def pause_mission(self) -> ControlMissionResponse:
        """Pause mission

        Raises
        ------
        HTTTPException 408 Request timeout
            If there is a timeout while communicating with the state machine
        """
        try:
            response = self._send_command(True, self.api_events.pause_mission)
            self.logger.info("OK - Mission successfully paused")
            return response
        except EventConflictError:
            error_message = "Previous pause mission request is still being processed"
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except EventTimeoutError:
            error_message = (
                "State machine has entered a state which cannot pause a mission"
            )
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)

    def resume_mission(self) -> ControlMissionResponse:
        """Resume mission

        Raises
        ------
        HTTTPException 408 Request timeout
            If there is a timeout while communicating with the state machine
        """
        try:
            response = self._send_command(True, self.api_events.resume_mission)
            self.logger.info("OK - Mission successfully resumed")
            return response
        except EventConflictError:
            error_message = "Previous resume mission request is still being processed"
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except EventTimeoutError:
            error_message = "Internal Server Error - Failed to resume mission"
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )

    def stop_mission(self, mission_id: str = "") -> ControlMissionResponse:
        """Stop mission

        Raises
        ------
        HTTPException 404 Not Found
            If the mission_id was not known to Isar
        HTTPException 503 Service Unavailable
            The request was understood, but attempting to stop the mission failed
        HTTPException 408 Request timeout
            If there is a timeout while communicating with the state machine
        """
        try:
            stop_mission_response: ControlMissionResponse = self._send_command(
                mission_id, self.api_events.stop_mission
            )

            if stop_mission_response.mission_not_found:
                error_message = f"Mission ID {stop_mission_response.mission_id} is not currently running"
                self.logger.error(error_message)
                raise HTTPException(
                    status_code=HTTPStatus.NOT_FOUND, detail=error_message
                )

            if stop_mission_response.mission_status != MissionStatus.Cancelled.value:
                error_message = "Failed to stop mission"
                self.logger.error(error_message)
                raise HTTPException(
                    status_code=HTTPStatus.CONFLICT, detail=error_message
                )
        except EventConflictError:
            error_message = "Previous stop mission request is still being processed"
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        except EventTimeoutError:
            error_message = (
                "State machine has entered a state which cannot stop a mission"
            )
            self.logger.warning(error_message)
            raise HTTPException(status_code=HTTPStatus.CONFLICT, detail=error_message)
        self.logger.info("OK - Mission successfully stopped")
        return stop_mission_response

    def release_intervention_needed(self) -> None:
        """Release intervention needed state

        Raises
        ------
        HTTPException 500 Internal Server Error
            If the intervention needed state could not be released
        """
        try:
            self._send_command(True, self.api_events.release_intervention_needed)
            self.logger.info("OK - Intervention needed state released")
        except EventTimeoutError:
            error_message = (
                "Internal Server Error - Failed to release intervention needed state"
            )
            self.logger.error(error_message)
            raise HTTPException(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=error_message
            )

    def _send_command(self, input: T1, api_event: APIEvent[T1, T2]) -> T2:
        if api_event.request.has_event() or api_event.response.has_event():
            raise EventConflictError("API event has already been sent")

        try:
            api_event.request.trigger_event(input, timeout=1)
            return api_event.response.consume_event(timeout=self.queue_timeout)
        except EventTimeoutError as e:
            self.logger.error("Queue timed out")
            api_event.request.clear_event()
            self.logger.error("No output received for command to state machine")
            raise e



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine

from isar.apis.models.models import MissionStartResponse
from robot_interface.models.exceptions.robot_exceptions import (
    ErrorMessage,
    RobotException,
    RobotInitializeException,
)
from robot_interface.models.mission.status import MissionStatus, TaskStatus


def acknowledge_mission(state_machine: "StateMachine") -> bool:
    state_machine.events.api_requests.start_mission.response.put(
        MissionStartResponse(mission_started=True)
    )
    return True


def prepare_state_machine_before_mission(state_machine: "StateMachine") -> bool:
    state_machine.logger.info(
        "Initiating mission:\n"
        f"  Mission ID: {state_machine.current_mission.id}\n"
        f"  Mission Name: {state_machine.current_mission.name}\n"
        f"  Number of Tasks: {len(state_machine.current_mission.tasks)}"
    )

    state_machine.current_mission.status = MissionStatus.InProgress
    state_machine.publish_mission_status()
    state_machine.current_task = state_machine.task_selector.next_task()
    state_machine.send_task_status()
    if state_machine.current_task is None:
        return False

    state_machine.current_task.status = TaskStatus.InProgress
    state_machine.publish_task_status(task=state_machine.current_task)
    return True


def initialize_robot(state_machine: "StateMachine") -> bool:
    try:
        state_machine.robot.initialize()
    except (RobotInitializeException, RobotException) as e:
        state_machine.current_task.error_message = ErrorMessage(
            error_reason=e.error_reason, error_description=e.error_description
        )
        state_machine.logger.error(
            f"Failed to initialize robot because: {e.error_description}"
        )
        _initialization_failed(state_machine)
        return False
    return True


def set_mission_to_in_progress(state_machine: "StateMachine") -> bool:
    state_machine.current_mission.status = MissionStatus.InProgress
    state_machine.publish_task_status(task=state_machine.current_task)
    state_machine.logger.info(
        f"Successfully initiated "
        f"{type(state_machine.current_task).__name__} "
        f"task: {str(state_machine.current_task.id)[:8]}"
    )
    return True


def trigger_start_mission_event(state_machine: "StateMachine") -> bool:
    state_machine.events.state_machine_events.start_mission.put(
        state_machine.current_mission
    )
    return True


def _initialization_failed(state_machine: "StateMachine") -> None:
    state_machine.events.api_requests.start_mission.response.put(
        MissionStartResponse(
            mission_started=False,
            mission_not_started_reason="Failed to initialize robot",
        )
    )
    state_machine._finalize()



from typing import TYPE_CHECKING, Callable, Optional

from isar.apis.models.models import ControlMissionResponse, MissionStartResponse
from isar.models.events import Event, EventTimeoutError
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import RobotStatus, TaskStatus

if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


def start_mission_event_handler(
    state_machine: "StateMachine",
    event: Event[Mission],
    response: Event[MissionStartResponse],
) -> Optional[Callable]:
    mission: Optional[Mission] = event.consume_event()
    if mission:
        if not state_machine.battery_level_is_above_mission_start_threshold():
            try:
                response.trigger_event(
                    MissionStartResponse(
                        mission_id=mission.id,
                        mission_started=False,
                        mission_not_started_reason="Robot battery too low",
                    ),
                    timeout=1,  # This conflict can happen if two API requests are received at the same time
                )
            except EventTimeoutError:
                pass
            return None
        state_machine.start_mission(mission=mission)
        return state_machine.request_mission_start  # type: ignore
    return None


def return_home_event_handler(
    state_machine: "StateMachine", event: Event[bool]
) -> Optional[Callable]:
    if event.consume_event():
        state_machine.events.api_requests.return_home.response.put(True)
        return state_machine.request_return_home  # type: ignore
    return None


def robot_status_event_handler(
    state_machine: "StateMachine",
    expected_status: RobotStatus,
    event: Event[RobotStatus],
) -> Optional[Callable]:
    robot_status: RobotStatus = event.check()
    if robot_status != expected_status:
        return state_machine.robot_status_changed  # type: ignore
    return None


def stop_mission_event_handler(
    state_machine: "StateMachine", event: Event[str]
) -> Optional[Callable]:
    mission_id: str = event.consume_event()
    if mission_id is not None:
        if state_machine.current_mission.id == mission_id or mission_id == "":
            return state_machine.stop  # type: ignore
        else:
            state_machine.events.api_requests.stop_mission.response.put(
                ControlMissionResponse(
                    mission_id=mission_id,
                    mission_status=state_machine.current_mission.status,
                    mission_not_found=True,
                    task_id=state_machine.current_task.id,
                    task_status=state_machine.current_task.status,
                )
            )
    return None


def mission_started_event_handler(
    state_machine: "StateMachine",
    event: Event[bool],
) -> Optional[Callable]:
    if event.consume_event():
        state_machine.mission_ongoing = True
    return None


def mission_failed_event_handler(
    state_machine: "StateMachine",
    event: Event[Optional[ErrorMessage]],
) -> Optional[Callable]:
    mission_failed: Optional[ErrorMessage] = event.consume_event()
    if mission_failed is not None:
        state_machine.logger.warning(
            f"Failed to initiate mission "
            f"{str(state_machine.current_mission.id)[:8]} because: "
            f"{mission_failed.error_description}"
        )
        state_machine.current_mission.error_message = ErrorMessage(
            error_reason=mission_failed.error_reason,
            error_description=mission_failed.error_description,
        )
        return state_machine.mission_failed_to_start  # type: ignore
    return None


def task_status_failed_event_handler(
    state_machine: "StateMachine",
    handle_task_completed: Callable[[TaskStatus], Callable],
    event: Event[Optional[ErrorMessage]],
) -> Optional[Callable]:
    if not state_machine.mission_ongoing:
        return None

    task_failure: Optional[ErrorMessage] = event.consume_event()
    if task_failure is not None:
        if state_machine.current_task is None:
            state_machine.logger.warning(
                "Received task status failed event when no task was running"
            )
            return None
        state_machine.awaiting_task_status = False
        state_machine.current_task.error_message = task_failure
        state_machine.logger.error(
            f"Monitoring task {state_machine.current_task.id[:8]} failed "
            f"because: {task_failure.error_description}"
        )
        return _handle_new_task_status(
            state_machine, handle_task_completed, TaskStatus.Failed
        )

    elif (
        not state_machine.awaiting_task_status
        and state_machine.current_task is not None
    ):
        state_machine.events.state_machine_events.task_status_request.trigger_event(
            state_machine.current_task.id,
        )
        state_machine.awaiting_task_status = True
    return None


def task_status_event_handler(
    state_machine: "StateMachine",
    handle_task_completed: Callable[[TaskStatus], Callable],
    event: Event[Optional[TaskStatus]],
) -> Optional[Callable]:
    if not state_machine.mission_ongoing:
        return None

    status: Optional[TaskStatus] = event.consume_event()
    if status is not None:
        state_machine.awaiting_task_status = False
        return _handle_new_task_status(state_machine, handle_task_completed, status)

    elif (
        not state_machine.awaiting_task_status
        and state_machine.current_task is not None
    ):
        state_machine.events.state_machine_events.task_status_request.trigger_event(
            state_machine.current_task.id,
        )
        state_machine.awaiting_task_status = True
    return None


def _handle_new_task_status(
    state_machine: "StateMachine",
    handle_task_completed: Callable[[TaskStatus], Callable],
    status: TaskStatus,
) -> Optional[Callable]:
    if state_machine.current_task is None:
        state_machine.iterate_current_task()

    state_machine.current_task.status = status

    if state_machine.current_task.is_finished():
        state_machine.report_task_status(state_machine.current_task)
        state_machine.publish_task_status(task=state_machine.current_task)

        return handle_task_completed(status)
    return None



import logging
from threading import Event, Thread

from isar.config.settings import settings
from isar.models.events import RobotServiceEvents
from robot_interface.models.exceptions.robot_exceptions import (
    ErrorMessage,
    RobotException,
    RobotInfeasibleMissionException,
)
from robot_interface.models.mission.mission import Mission
from robot_interface.robot_interface import RobotInterface


class RobotStartMissionThread(Thread):
    def __init__(
        self,
        robot_service_events: RobotServiceEvents,
        robot: RobotInterface,
        signal_thread_quitting: Event,
        mission: Mission,
    ):
        self.logger = logging.getLogger("robot")
        self.robot_service_events: RobotServiceEvents = robot_service_events
        self.robot: RobotInterface = robot
        self.signal_thread_quitting: Event = signal_thread_quitting
        self.mission = mission
        Thread.__init__(self, name="Robot start mission thread")

    def run(self):
        retries = 0
        started_mission = False
        try:
            while not started_mission:
                if self.signal_thread_quitting.wait(0):
                    return
                try:
                    self.robot.initiate_mission(self.mission)
                except RobotInfeasibleMissionException as e:
                    self.logger.error(
                        f"Mission is infeasible and cannot be scheduled because: {e.error_description}"
                    )
                    self.robot_service_events.mission_failed.trigger_event(
                        ErrorMessage(
                            error_reason=e.error_reason,
                            error_description=e.error_description,
                        )
                    )

                    break
                except RobotException as e:
                    retries += 1
                    self.logger.warning(
                        f"Initiating failed #: {str(retries)} "
                        f"because: {e.error_description}"
                    )

                    if retries >= settings.INITIATE_FAILURE_COUNTER_LIMIT:
                        self.logger.error(
                            f"Mission will be cancelled after failing to initiate "
                            f"{settings.INITIATE_FAILURE_COUNTER_LIMIT} times because: "
                            f"{e.error_description}"
                        )

                        self.robot_service_events.mission_failed.trigger_event(
                            ErrorMessage(
                                error_reason=e.error_reason,
                                error_description=e.error_description,
                            )
                        )
                        break

                    continue

                started_mission = True
        except RobotInfeasibleMissionException as e:
            self.robot_service_events.mission_failed.trigger_event(
                ErrorMessage(
                    error_reason=e.error_reason, error_description=e.error_description
                ),
            )

        if started_mission:
            self.robot_service_events.mission_started.trigger_event(True)



import logging
from threading import Event as ThreadEvent
from typing import Optional

from isar.models.events import (
    Event,
    Events,
    RobotServiceEvents,
    SharedState,
    StateMachineEvents,
)
from isar.robot.robot_start_mission import RobotStartMissionThread
from isar.robot.robot_status import RobotStatusThread
from isar.robot.robot_stop_mission import RobotStopMissionThread
from isar.robot.robot_task_status import RobotTaskStatusThread
from robot_interface.models.exceptions.robot_exceptions import ErrorMessage, ErrorReason
from robot_interface.models.mission.mission import Mission
from robot_interface.robot_interface import RobotInterface


class Robot(object):
    def __init__(
        self, events: Events, robot: RobotInterface, shared_state: SharedState
    ) -> None:
        self.logger = logging.getLogger("robot")
        self.state_machine_events: StateMachineEvents = events.state_machine_events
        self.robot_service_events: RobotServiceEvents = events.robot_service_events
        self.shared_state: SharedState = shared_state
        self.robot: RobotInterface = robot
        self.start_mission_thread: Optional[RobotStartMissionThread] = None
        self.robot_status_thread: Optional[RobotStatusThread] = None
        self.robot_task_status_thread: Optional[RobotTaskStatusThread] = None
        self.stop_mission_thread: Optional[RobotStopMissionThread] = None
        self.signal_thread_quitting: ThreadEvent = ThreadEvent()

    def stop(self) -> None:
        self.signal_thread_quitting.set()
        if self.robot_status_thread is not None and self.robot_status_thread.is_alive():
            self.robot_status_thread.join()
        if (
            self.robot_task_status_thread is not None
            and self.robot_task_status_thread.is_alive()
        ):
            self.robot_task_status_thread.join()
        if (
            self.start_mission_thread is not None
            and self.start_mission_thread.is_alive()
        ):
            self.start_mission_thread.join()
        if self.stop_mission_thread is not None and self.stop_mission_thread.is_alive():
            self.stop_mission_thread.join()
        self.robot_status_thread = None
        self.robot_task_status_thread = None
        self.start_mission_thread = None

    def _start_mission_event_handler(self, event: Event[Mission]) -> None:
        start_mission = event.consume_event()
        if start_mission is not None:
            if (
                self.start_mission_thread is not None
                and self.start_mission_thread.is_alive()
            ):
                self.logger.warning(
                    "Attempted to start mission while another mission was starting."
                )
                self.start_mission_thread.join()
            self.start_mission_thread = RobotStartMissionThread(
                self.robot_service_events,
                self.robot,
                self.signal_thread_quitting,
                start_mission,
            )
            self.start_mission_thread.start()

    def _task_status_request_handler(self, event: Event[str]) -> None:
        task_id: str = event.consume_event()
        if task_id:
            self.robot_task_status_thread = RobotTaskStatusThread(
                self.robot_service_events,
                self.robot,
                self.signal_thread_quitting,
                task_id,
            )
            self.robot_task_status_thread.start()

    def _stop_mission_request_handler(self, event: Event[bool]) -> None:
        if event.consume_event():
            if (
                self.stop_mission_thread is not None
                and self.stop_mission_thread.is_alive()
            ):
                self.logger.warning(
                    "Received stop mission event while trying to stop a mission. Aborting stop attempt."
                )
                return
            if (
                self.start_mission_thread is not None
                and self.start_mission_thread.is_alive()
            ):
                error_description = "Received stop mission event while trying to start a mission. Aborting stop attempt."
                error_message = ErrorMessage(
                    error_reason=ErrorReason.RobotStillStartingMissionException,
                    error_description=error_description,
                )
                self.robot_service_events.mission_failed_to_stop.trigger_event(
                    error_message
                )
                return
            self.stop_mission_thread = RobotStopMissionThread(
                self.robot_service_events, self.robot, self.signal_thread_quitting
            )
            self.stop_mission_thread.start()

    def run(self) -> None:
        self.robot_status_thread = RobotStatusThread(
            self.robot, self.signal_thread_quitting, self.shared_state
        )
        self.robot_status_thread.start()

        while not self.signal_thread_quitting.wait(0):
            self._start_mission_event_handler(self.state_machine_events.start_mission)

            self._task_status_request_handler(
                self.state_machine_events.task_status_request
            )

            self._stop_mission_request_handler(self.state_machine_events.stop_mission)

        self.logger.info("Exiting robot service main thread")



import logging
import time
from threading import Event, Thread
from typing import Optional

from isar.config.settings import settings
from isar.models.events import RobotServiceEvents
from robot_interface.models.exceptions.robot_exceptions import (
    ErrorMessage,
    RobotActionException,
    RobotException,
)
from robot_interface.robot_interface import RobotInterface


class RobotStopMissionThread(Thread):
    def __init__(
        self,
        robot_service_events: RobotServiceEvents,
        robot: RobotInterface,
        signal_thread_quitting: Event,
    ):
        self.logger = logging.getLogger("robot")
        self.robot_service_events: RobotServiceEvents = robot_service_events
        self.robot: RobotInterface = robot
        self.signal_thread_quitting: Event = signal_thread_quitting
        Thread.__init__(self, name="Robot start mission thread")

    def run(self) -> None:
        retries = 0
        error: Optional[ErrorMessage] = None
        while retries < settings.STOP_ROBOT_ATTEMPTS_LIMIT:
            if self.signal_thread_quitting.wait(0):
                return

            try:
                self.robot.stop()
            except (RobotActionException, RobotException) as e:
                self.logger.warning(
                    f"\nFailed to stop robot because: {e.error_description}"
                    f"\nAttempting to stop the robot again"
                )
                retries += 1
                error = ErrorMessage(
                    error_reason=e.error_reason, error_description=e.error_description
                )
                time.sleep(settings.FSM_SLEEP_TIME)
                continue

            self.robot_service_events.mission_successfully_stopped.trigger_event(True)
            return

        error_description = (
            f"\nFailed to stop the robot after {retries} attempts because: "
            f"{error.error_description}"
            f"\nBe aware that the robot may still be moving even though a stop has "
            "been attempted"
        )

        error_message = ErrorMessage(
            error_reason=error.error_reason,
            error_description=error_description,
        )

        self.robot_service_events.mission_failed_to_stop.trigger_event(error_message)



import json
import logging
from collections import deque
from datetime import datetime, timezone
from threading import Event
from typing import Deque, List, Optional, Tuple

from transitions import Machine
from transitions.core import State

from isar.apis.models.models import ControlMissionResponse
from isar.config.settings import settings
from isar.mission_planner.task_selector_interface import (
    TaskSelectorInterface,
    TaskSelectorStop,
)
from isar.models.events import Events, SharedState
from isar.state_machine.states.await_next_mission import AwaitNextMission
from isar.state_machine.states.blocked_protective_stop import BlockedProtectiveStop
from isar.state_machine.states.home import Home
from isar.state_machine.states.intervention_needed import InterventionNeeded
from isar.state_machine.states.monitor import Monitor
from isar.state_machine.states.offline import Offline
from isar.state_machine.states.paused import Paused
from isar.state_machine.states.recharging import Recharging
from isar.state_machine.states.returning_home import ReturningHome
from isar.state_machine.states.robot_standing_still import RobotStandingStill
from isar.state_machine.states.stopping import Stopping
from isar.state_machine.states.unknown_status import UnknownStatus
from isar.state_machine.states_enum import States
from isar.state_machine.transitions.mission import get_mission_transitions
from isar.state_machine.transitions.return_home import get_return_home_transitions
from isar.state_machine.transitions.robot_status import get_robot_status_transitions
from robot_interface.models.exceptions.robot_exceptions import (
    ErrorMessage,
    RobotException,
    RobotRetrieveInspectionException,
)
from robot_interface.models.inspection.inspection import Inspection
from robot_interface.models.mission.mission import Mission
from robot_interface.models.mission.status import RobotStatus, TaskStatus
from robot_interface.models.mission.task import TASKS, InspectionTask, Task
from robot_interface.robot_interface import RobotInterface
from robot_interface.telemetry.mqtt_client import MqttClientInterface
from robot_interface.telemetry.payloads import (
    InterventionNeededPayload,
    MissionAbortedPayload,
    MissionPayload,
    RobotStatusPayload,
    TaskPayload,
)
from robot_interface.utilities.json_service import EnhancedJSONEncoder


class StateMachine(object):
    """Handles state transitions for supervisory robot control."""

    def __init__(
        self,
        events: Events,
        shared_state: SharedState,
        robot: RobotInterface,
        mqtt_publisher: MqttClientInterface,
        task_selector: TaskSelectorInterface,
        sleep_time: float = settings.FSM_SLEEP_TIME,
        stop_robot_attempts_limit: int = settings.STOP_ROBOT_ATTEMPTS_LIMIT,
        transitions_log_length: int = settings.STATE_TRANSITIONS_LOG_LENGTH,
    ):
        """Initializes the state machine.

        Parameters
        ----------
        events : Events
            Events used for API and robot service communication.
        robot : RobotInterface
            Instance of robot interface.
        mqtt_publisher : MqttClientInterface
            Instance of MQTT client interface which has a publish function
        sleep_time : float
            Time to sleep in between state machine iterations.
        stop_robot_attempts_limit : int
            Maximum attempts to stop the robot when stop command is received
        transitions_log_length : int
            Length of state transition log list.

        """
        self.logger = logging.getLogger("state_machine")

        self.events: Events = events
        self.shared_state: SharedState = shared_state
        self.robot: RobotInterface = robot
        self.mqtt_publisher: Optional[MqttClientInterface] = mqtt_publisher
        self.task_selector: TaskSelectorInterface = task_selector

        self.signal_state_machine_to_stop: Event = Event()

        # List of states
        # States running mission
        self.monitor_state: State = Monitor(self)
        self.returning_home_state: State = ReturningHome(self)
        self.stopping_state: State = Stopping(self)
        self.paused_state: State = Paused(self)

        # States Waiting for mission
        self.await_next_mission_state: State = AwaitNextMission(self)
        self.home_state: State = Home(self)
        self.robot_standing_still_state: State = RobotStandingStill(self)
        self.intervention_needed_state: State = InterventionNeeded(self)

        # Status states
        self.offline_state: State = Offline(self)
        self.blocked_protective_stopping_state: State = BlockedProtectiveStop(self)
        self.recharging_state: State = Recharging(self)

        # Error and special status states
        self.unknown_status_state: State = UnknownStatus(self)

        self.states: List[State] = [
            self.monitor_state,
            self.returning_home_state,
            self.stopping_state,
            self.paused_state,
            self.await_next_mission_state,
            self.home_state,
            self.robot_standing_still_state,
            self.offline_state,
            self.blocked_protective_stopping_state,
            self.unknown_status_state,
            self.intervention_needed_state,
            self.recharging_state,
        ]

        self.machine = Machine(
            self, states=self.states, initial="unknown_status", queued=True
        )

        self.transitions: List[dict] = []

        self.transitions.extend(get_mission_transitions(self))
        self.transitions.extend(get_return_home_transitions(self))
        self.transitions.extend(get_robot_status_transitions(self))

        self.machine.add_transitions(self.transitions)

        self.stop_robot_attempts_limit: int = stop_robot_attempts_limit
        self.sleep_time: float = sleep_time

        self.current_mission: Optional[Mission] = None
        self.current_task: Optional[TASKS] = None

        self.mission_ongoing: bool = False

        self.current_state: State = States(self.state)  # type: ignore

        self.awaiting_task_status: bool = False

        self.transitions_log_length: int = transitions_log_length
        self.transitions_list: Deque[States] = deque([], self.transitions_log_length)

    #################################################################################

    def _finalize(self) -> None:
        self.publish_mission_status()
        self.log_mission_overview(mission=self.current_mission)
        state_transitions: str = ", ".join(
            [
                f"\n  {transition}" if (i + 1) % 10 == 0 else f"{transition}"
                for i, transition in enumerate(list(self.transitions_list))
            ]
        )
        self.logger.info("State transitions:\n  %s", state_transitions)
        self.reset_state_machine()

    def begin(self):
        """Starts the state machine. Transitions into unknown status state."""
        self.robot_status_changed()  # type: ignore

    def terminate(self):
        self.logger.info("Stopping state machine")
        self.signal_state_machine_to_stop.set()

    def iterate_current_task(self):
        if self.current_task is None:
            raise ValueError("No current task is set")

        if self.current_task.is_finished():
            try:
                self.current_task = self.task_selector.next_task()
                self.current_task.status = TaskStatus.InProgress
                self.publish_task_status(task=self.current_task)
            except TaskSelectorStop:
                # Indicates that all tasks are finished
                self.current_task = None
            self.send_task_status()

    def battery_level_is_above_mission_start_threshold(self):
        return (
            not self.shared_state.robot_battery_level.check()
            < settings.ROBOT_MISSION_BATTERY_START_THRESHOLD
        )

    def update_state(self):
        """Updates the current state of the state machine."""
        self.current_state = States(self.state)  # type: ignore
        self.shared_state.state.update(self.current_state)
        self._log_state_transition(self.current_state)
        self.logger.info("State: %s", self.current_state)
        self.publish_status()

    def reset_state_machine(self) -> None:
        self.logger.info("Resetting state machine")
        self.current_task = None
        self.send_task_status()
        self.current_mission = None
        self.mission_ongoing = False

    def start_mission(self, mission: Mission):
        """Starts a scheduled mission."""
        self.current_mission = mission

        self.task_selector.initialize(tasks=self.current_mission.tasks)

    def send_task_status(self):
        self.shared_state.state_machine_current_task.update(self.current_task)

    def report_task_status(self, task: Task) -> None:
        if task.status == TaskStatus.Failed:
            self.logger.warning(
                f"Task: {str(task.id)[:8]} was reported as failed by the robot"
            )
        elif task.status == TaskStatus.Successful:
            self.logger.info(
                f"{type(task).__name__} task: {str(task.id)[:8]} completed"
            )
        else:
            self.logger.info(
                f"Task: {str(task.id)[:8]} was reported as task.status by the robot"
            )

    def publish_mission_aborted(self, reason: str, can_be_continued: bool) -> None:
        if not self.mqtt_publisher:
            return

        if self.current_mission is None:
            self.logger.warning(
                "Could not publish mission aborted message. No ongoing mission."
            )
            return

        payload: MissionAbortedPayload = MissionAbortedPayload(
            isar_id=settings.ISAR_ID,
            robot_name=settings.ROBOT_NAME,
            mission_id=self.current_mission.id,
            reason=reason,
            can_be_continued=can_be_continued,
            timestamp=datetime.now(timezone.utc),
        )

        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_MISSION_ABORTED,
            payload=json.dumps(payload, cls=EnhancedJSONEncoder),
            qos=1,
            retain=True,
        )

    def publish_mission_status(self) -> None:
        if not self.mqtt_publisher:
            return

        error_message: Optional[ErrorMessage] = None
        if self.current_mission:
            if self.current_mission.error_message:
                error_message = self.current_mission.error_message

        payload: MissionPayload = MissionPayload(
            isar_id=settings.ISAR_ID,
            robot_name=settings.ROBOT_NAME,
            mission_id=self.current_mission.id if self.current_mission else None,
            status=self.current_mission.status if self.current_mission else None,
            error_reason=error_message.error_reason if error_message else None,
            error_description=(
                error_message.error_description if error_message else None
            ),
            timestamp=datetime.now(timezone.utc),
        )

        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_MISSION,
            payload=json.dumps(payload, cls=EnhancedJSONEncoder),
            qos=1,
            retain=True,
        )

    def publish_task_status(self, task: TASKS) -> None:
        """Publishes the task status to the MQTT Broker"""
        if not self.mqtt_publisher:
            return

        error_message: Optional[ErrorMessage] = None
        if task:
            if task.error_message:
                error_message = task.error_message

        payload: TaskPayload = TaskPayload(
            isar_id=settings.ISAR_ID,
            robot_name=settings.ROBOT_NAME,
            mission_id=self.current_mission.id if self.current_mission else None,
            task_id=task.id if task else None,
            status=task.status if task else None,
            task_type=task.type if task else None,
            error_reason=error_message.error_reason if error_message else None,
            error_description=(
                error_message.error_description if error_message else None
            ),
            timestamp=datetime.now(timezone.utc),
        )

        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_TASK,
            payload=json.dumps(payload, cls=EnhancedJSONEncoder),
            qos=1,
            retain=True,
        )

    def publish_intervention_needed(self, error_message: str) -> None:
        """Publishes the intervention needed message to the MQTT Broker"""
        if not self.mqtt_publisher:
            return

        payload: InterventionNeededPayload = InterventionNeededPayload(
            isar_id=settings.ISAR_ID,
            robot_name=settings.ROBOT_NAME,
            reason=error_message,
            timestamp=datetime.now(timezone.utc),
        )

        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_INTERVENTION_NEEDED,
            payload=json.dumps(payload, cls=EnhancedJSONEncoder),
            qos=1,
            retain=True,
        )

    def publish_status(self) -> None:
        if not self.mqtt_publisher:
            return

        payload: RobotStatusPayload = RobotStatusPayload(
            isar_id=settings.ISAR_ID,
            robot_name=settings.ROBOT_NAME,
            status=self._current_status(),
            timestamp=datetime.now(timezone.utc),
        )

        self.mqtt_publisher.publish(
            topic=settings.TOPIC_ISAR_STATUS,
            payload=json.dumps(payload, cls=EnhancedJSONEncoder),
            qos=1,
            retain=True,
        )

    def _current_status(self) -> RobotStatus:
        if (
            self.current_state == States.RobotStandingStill
            or self.current_state == States.AwaitNextMission
        ):
            return RobotStatus.Available
        elif self.current_state == States.Home:
            return RobotStatus.Home
        elif self.current_state == States.ReturningHome:
            return RobotStatus.ReturningHome
        elif self.current_state == States.Offline:
            return RobotStatus.Offline
        elif self.current_state == States.BlockedProtectiveStop:
            return RobotStatus.BlockedProtectiveStop
        elif self.current_state == States.InterventionNeeded:
            return RobotStatus.InterventionNeeded
        elif self.current_state == States.Recharging:
            return RobotStatus.Recharging
        else:
            return RobotStatus.Busy

    def _log_state_transition(self, next_state) -> None:
        """Logs all state transitions that are not self-transitions."""
        self.transitions_list.append(next_state)

    def log_mission_overview(self, mission: Mission) -> None:
        """Log an overview of the tasks in a mission"""
        log_statements: List[str] = []
        for task in mission.tasks:
            log_statements.append(
                f"{type(task).__name__:<20} {str(task.id)[:8]:<32} -- {task.status}"
            )
        log_statement: str = "\n".join(log_statements)

        self.logger.info("Mission overview:\n%s", log_statement)

    def _make_control_mission_response(self) -> ControlMissionResponse:
        return ControlMissionResponse(
            mission_id=self.current_mission.id if self.current_mission else None,
            mission_status=(
                self.current_mission.status if self.current_mission else None
            ),
            task_id=self.current_task.id if self.current_task else None,
            task_status=self.current_task.status if self.current_task else None,
        )

    def _queue_empty_response(self) -> None:
        self.events.api_requests.stop_mission.response.put(
            ControlMissionResponse(
                mission_id="None",
                mission_status="None",
                task_id="None",
                task_status="None",
            )
        )

    def should_upload_inspections(self) -> bool:
        if settings.UPLOAD_INSPECTIONS_ASYNC:
            return False

        return (
            self.current_task.is_finished()
            and self.current_task.status == TaskStatus.Successful
            and isinstance(self.current_task, InspectionTask)
        )

    def queue_inspections_for_upload(
        self, mission: Mission, current_task: InspectionTask, logger: logging.Logger
    ) -> None:
        try:
            inspection: Inspection = self.robot.get_inspection(task=current_task)
            if current_task.inspection_id != inspection.id:
                logger.warning(
                    f"The inspection_id of task ({current_task.inspection_id}) "
                    f"and result ({inspection.id}) is not matching. "
                    f"This may lead to confusions when accessing the inspection later"
                )

        except (RobotRetrieveInspectionException, RobotException) as e:
            error_message: ErrorMessage = ErrorMessage(
                error_reason=e.error_reason, error_description=e.error_description
            )
            self.current_task.error_message = error_message
            logger.error(
                f"Failed to retrieve inspections because: {e.error_description}"
            )
            return

        except Exception as e:
            logger.error(
                f"Failed to retrieve inspections because of unexpected error: {e}"
            )
            return

        if not inspection:
            logger.warning(
                f"No inspection result data retrieved for task {str(current_task.id)[:8]}"
            )

        inspection.metadata.tag_id = current_task.tag_id

        message: Tuple[Inspection, Mission] = (
            inspection,
            mission,
        )
        self.events.upload_queue.put(message)
        logger.info(f"Inspection result: {str(inspection.id)[:8]} queued for upload")


def main(state_machine: StateMachine):
    """Starts a state machine instance."""
    state_machine.begin()



import logging
import time
from copy import deepcopy
from dataclasses import dataclass
from threading import Event as ThreadEvent
from typing import TYPE_CHECKING, Callable, Generic, List, Optional, TypeVar

from transitions import State

from isar.config.settings import settings
from isar.models.events import Event

T = TypeVar("T")


@dataclass
class EventHandlerMapping(Generic[T]):
    name: str
    event: Event[T]
    handler: Callable[[Event[T]], Optional[Callable]]


@dataclass
class TimeoutHandlerMapping:
    name: str
    timeout_in_seconds: float
    handler: Callable[[], Optional[Callable]]


if TYPE_CHECKING:
    from isar.state_machine.state_machine import StateMachine


class EventHandlerBase(State):
    def __init__(
        self,
        state_machine: "StateMachine",
        state_name: str,
        event_handler_mappings: List[EventHandlerMapping],
        timers: List[TimeoutHandlerMapping] = [],
    ) -> None:

        super().__init__(name=state_name, on_enter=self.start)
        self.state_machine: "StateMachine" = state_machine
        self.logger = logging.getLogger("state_machine")
        self.events = state_machine.events
        self.signal_state_machine_to_stop: ThreadEvent = (
            state_machine.signal_state_machine_to_stop
        )
        self.event_handler_mappings = event_handler_mappings
        self.state_name: str = state_name
        self.timers = timers

    def start(self) -> None:
        self.state_machine.update_state()
        self._run()

    def stop(self) -> None:
        return

    def get_event_handler_by_name(
        self, event_handler_name: str
    ) -> Optional[EventHandlerMapping]:
        filtered_handlers = list(
            filter(
                lambda mapping: mapping.name == event_handler_name,
                self.event_handler_mappings,
            )
        )
        return filtered_handlers[0] if len(filtered_handlers) > 0 else None

    def get_event_timer_by_name(
        self, event_timer_name: str
    ) -> Optional[TimeoutHandlerMapping]:
        filtered_timers = list(
            filter(
                lambda mapping: mapping.name == event_timer_name,
                self.timers,
            )
        )
        return filtered_timers[0] if len(filtered_timers) > 0 else None

    def _run(self) -> None:
        should_exit_state: bool = False
        timers = deepcopy(self.timers)
        entered_time = time.time()
        while True:
            if self.signal_state_machine_to_stop.is_set():
                self.logger.info(
                    "Stopping state machine from %s state", self.state_name
                )
                break

            for timer in timers:
                if time.time() - entered_time > timer.timeout_in_seconds:
                    transition_func = timer.handler()
                    timers.remove(timer)
                    if transition_func is not None:
                        transition_func()
                        should_exit_state = True
                        break

            if should_exit_state:
                break

            for handler_mapping in self.event_handler_mappings:
                transition_func = handler_mapping.handler(handler_mapping.event)
                if transition_func is not None:
                    transition_func()
                    should_exit_state = True
                    break

            if should_exit_state:
                break
            time.sleep(settings.FSM_SLEEP_TIME)
